from __future__ import annotations

from typing import Literal, overload

from ibis.expr import types as ir

from mismo._util import bind_one, cases
from mismo.arrays import array_combinations, array_min
from mismo.compare import MatchLevel
from mismo.text import damerau_levenshtein


@overload
def clean_phone_number(
    phones: ir.Table, *, default_area_code: str | None = None
) -> ir.Table: ...


@overload
def clean_phone_number(
    phones: ir.ArrayValue, *, default_area_code: str | None = None
) -> ir.ArrayValue: ...


@overload
def clean_phone_number(
    phones: ir.StringValue, *, default_area_code: str | None = None
) -> ir.StringValue: ...


def clean_phone_number(numbers, *, default_area_code=None):
    """Extracts any 10-digit number from a string.

    Drops leading 1 country code if present.

    Parsing failures are returned as NULL.

    Empty strings are returned as NULL.

    If you supply a default_area_code, it will be prepended to 7-digit numbers.

    If a number looks bogus, ie it contains "0000", "9999", or "12345",
    it is set to NULL.
    """

    def f(n):
        return _clean_phone_number(n, default_area_code=default_area_code)

    if isinstance(numbers, ir.StringValue):
        return f(numbers)
    elif isinstance(numbers, ir.ArrayValue):
        return numbers.map(f).filter(lambda x: x.notnull()).unique()
    elif isinstance(numbers, ir.Table):
        return numbers.mutate(
            phones=clean_phone_number(
                numbers.phones, default_area_code=default_area_code
            )
        )
    raise ValueError(f"Unexpected type {type(numbers)}")


def _clean_phone_number(
    numbers: ir.StringValue, *, default_area_code: str | None = None
) -> ir.StringValue:
    x = numbers
    x = x.cast("string")
    x = x.re_replace(r"[^0-9]", "")
    if default_area_code:
        if len(default_area_code) != 3:
            raise ValueError("default_area_code must be 3 digits")
        x = x.re_replace(r"^(\d{7})$", rf"{default_area_code}\1")
    x = x.re_extract(r"1?(\d{10})", 1)
    x = x.nullif("")
    x = _drop_bogus_numbers(x)
    return x


def _drop_bogus_numbers(numbers: ir.StringValue) -> ir.StringValue:
    bogus_substrings = ["0000", "9999", "12345"]
    pattern = "|".join(bogus_substrings)
    is_bogus = numbers.re_search(".*" + pattern + ".*")
    return is_bogus.ifelse(None, numbers)


class PhoneMatchLevel(MatchLevel):
    """How closely two phone numbers match."""

    EXACT = 0
    """The numbers are exactly the same."""
    NEAR = 1
    """The numbers have a small edit distance."""
    ELSE = 2
    """None of the above."""


def match_level(
    p1: ir.StringValue,
    p2: ir.StringValue,
    *,
    native_representation: Literal["integer", "string"] = "integer",
) -> PhoneMatchLevel:
    """Match level of two phone numbers.

    Assumes the phone numbers have already been cleaned and normalized.

    Parameters
    ----------
    p1 :
        The first phone number.
    p2 :
        The second phone number.

    Returns
    -------
    level:
        The match level.
    """

    def f(level: MatchLevel):
        if native_representation == "string":
            return level.as_string()
        else:
            return level.as_integer()

    raw = cases(
        (p1 == p2, f(PhoneMatchLevel.EXACT)),
        (damerau_levenshtein(p1, p2) <= 1, f(PhoneMatchLevel.NEAR)),
        else_=f(PhoneMatchLevel.ELSE),
    )
    return PhoneMatchLevel(raw)


class PhonesDimension:
    """Prepares, blocks, and compares sets of phone numbers.

    This is useful if each record contains a collection of phone numbers.
    Two records are probably the same if they have a lot of phone numbers in common.
    """

    def __init__(
        self,
        column: str,
        *,
        column_cleaned: str = "{column}_cleaned",
        column_compared: str = "{column}_compared",
    ):
        """Initialize the dimension.

        Parameters
        ----------
        column :
            The name of the column that holds a array<string> of phone numbers.
        column_cleaned :
            The name of the column that will be filled with the parsed phone numbers.
        column_compared :
            The name of the column that will be filled with the comparison results.
        """
        self.column = column
        self.column_cleaned = column_cleaned.format(column=column)
        self.column_compared = column_compared.format(column=column)

    def prepare_for_fast_linking(self, t: ir.Table) -> ir.Table:
        """Add a column with the parsed and normalized phone numbers."""
        return t.mutate(
            bind_one(t, self.column).map(clean_phone_number).name(self.column_cleaned)
        )

    def prepare_for_blocking(self, t: ir.Table) -> ir.Table:
        """noop"""
        return t

    def compare(self, t: ir.Table) -> ir.Table:
        """Add a column with the best match between all pairs of phone numbers."""
        le = t[self.column_cleaned + "_l"]
        ri = t[self.column_cleaned + "_r"]
        pairs = array_combinations(le, ri)
        min_level = array_min(
            pairs.map(lambda pair: match_level(pair.l, pair.r).as_integer())
        ).fill_null(PhoneMatchLevel.ELSE.as_integer())
        return t.mutate(min_level.name(self.column_compared))



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

import ibis
from ibis import _

from mismo import _typing, _util, joins
from mismo.types._wrapper import TableWrapper

if TYPE_CHECKING:
    from mismo.types._linked_table import LinkedTable


class LinksTable(TableWrapper):
    """A table of links between two tables.

    This acts like an ibis table, guaranteed to have at least the
    columns record_id_l and record_id_r.
    It may have more columns, such as `address_match_level`,
    that describe the relationship between two records.

    In addition to the columns, this table has two properties, `left_` and `right_`,
    each of which is a [LinkedTable][mismo.LinkedTable] object,
    which is a wrapper around the left and right tables respectively.
    """

    # TODO:
    # Need to define whether or not it is allowed to have more than
    # one link between the same pair of records.

    record_id_l: ibis.Column
    """The record_id of the left table."""
    record_id_r: ibis.Column
    """The record_id of the right table."""

    def __init__(
        self, links: ibis.Table, *, left: ibis.Table, right: ibis.Table
    ) -> None:
        if "record_id_l" not in links.columns:
            raise ValueError("column 'record_id_l' not in links")
        if "record_id_r" not in links.columns:
            raise ValueError("column 'record_id_r' not in links")
        super().__init__(links)
        object.__setattr__(self, "_left_raw", left)
        object.__setattr__(self, "_right_raw", right)

    @classmethod
    def from_join_condition(
        cls,
        left: ibis.Table,
        right: ibis.Table,
        condition: Any,
    ) -> _typing.Self:
        if "record_id" not in left.columns:
            left = left.mutate(record_id=ibis.row_number())
        if "record_id" not in right.columns:
            right = right.mutate(record_id=ibis.row_number())
        links_raw = joins.join(
            left,
            right,
            condition,
            lname="{name}_l",
            rname="{name}_r",
            rename_all=True,
        )
        return LinksTable(links_raw, left=left, right=right)

    def with_left(
        self,
        *values: ibis.Deferred | Callable[[ibis.Table], ibis.Value] | None,
        **named_values: ibis.Deferred | Callable[[ibis.Table], ibis.Value] | None,
    ) -> LinksTable:
        """Add columns from the left table to this table of links.

        This allows you to add specific columns from the left table,
        renaming or modifying them as needed, following the `ibis.Table` pattern of
        `my_table.select("my_col", new_col=my_table.foo)`,
        except here we choose from the left table.

        Parameters
        ----------
        values
            The columns to add from the left table.
            Support string names, Deferreds, etc, just like `ibis.Table.select`.
        named_values
            Like values, but with names, just like `ibis.Table.select`.

        Examples
        --------
        >>> ibis.options.interactive = True
        >>> left = ibis.memtable({"record_id": [1, 2, 3], "address": ["a", "b", "c"]})
        >>> right = ibis.memtable({"record_id": [8, 9], "address": ["x", "y"]})
        >>> links_raw = ibis.memtable({"record_id_l": [1, 3], "record_id_r": [8, 9]})
        >>> links = LinksTable(links_raw, left=left, right=right)
        >>> links.with_left(
        ...     "address",
        ...     ibis._.address.upper().name("address_upper"),
        ...     left_address=ibis._.address,
        ... )
        ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━┓
        ┃ record_id_l ┃ record_id_r ┃ address ┃ address_upper ┃ left_address ┃
        ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━┩
        │ int64       │ int64       │ string  │ string        │ string       │
        ├─────────────┼─────────────┼─────────┼───────────────┼──────────────┤
        │           1 │           8 │ a       │ A             │ a            │
        │           3 │           9 │ c       │ C             │ c            │
        └─────────────┴─────────────┴─────────┴───────────────┴──────────────┘
        """
        if not values and not named_values:
            values = self.left.columns

        uname = _util.unique_name()
        left = self.left.select(_.record_id.name(uname), *values, **named_values)
        conflicts = [c for c in left.columns if c in self.columns]
        if conflicts:
            raise ValueError(f"conflicting columns: {conflicts}")
        joined = self.left_join(left, self.record_id_l == left[uname]).drop(uname)
        return LinksTable(joined, left=self._left_raw, right=self._right_raw)

    def with_right(
        self,
        *values: ibis.Deferred | Callable[[ibis.Table], ibis.Value] | None,
        **named_values: ibis.Deferred | Callable[[ibis.Table], ibis.Value] | None,
    ) -> LinksTable:
        """Add columns from the right table to this table of links.

        This allows you to add specific columns from the right table,
        renaming or modifying them as needed, following the `ibis.Table` pattern of
        `my_table.select("my_col", new_col=my_table.foo)`,
        except here we choose from the right table.

        Parameters
        ----------
        values
            The columns to add from the right table.
            Support string names, Deferreds, etc, just like `ibis.Table.select`.
        named_values
            Like values, but with names, just like `ibis.Table.select`.

        Examples
        --------
        >>> ibis.options.interactive = True
        >>> left = ibis.memtable({"record_id": [1, 2, 3], "address": ["a", "b", "c"]})
        >>> right = ibis.memtable({"record_id": [8, 9], "address": ["x", "y"]})
        >>> links_raw = ibis.memtable({"record_id_l": [1, 3], "record_id_r": [8, 9]})
        >>> links = LinksTable(links_raw, left=left, right=right)
        >>> links.with_right(
        ...     "address",
        ...     ibis._.address.upper().name("address_upper"),
        ...     right_address=ibis._.address,
        ... )
        ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┓
        ┃ record_id_l ┃ record_id_r ┃ address ┃ address_upper ┃ right_address ┃
        ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━┩
        │ int64       │ int64       │ string  │ string        │ string        │
        ├─────────────┼─────────────┼─────────┼───────────────┼───────────────┤
        │           1 │           8 │ x       │ X             │ x             │
        │           3 │           9 │ y       │ Y             │ y             │
        └─────────────┴─────────────┴─────────┴───────────────┴───────────────┘
        """
        if not values and not named_values:
            values = self.right.columns

        uname = _util.unique_name()
        right = self.right.select(_.record_id.name(uname), *values, **named_values)
        conflicts = [c for c in right.columns if c in self.columns]
        if conflicts:
            raise ValueError(f"conflicting columns: {conflicts}")
        joined = self.left_join(right, self.record_id_r == right[uname]).drop(uname)
        return LinksTable(joined, left=self._left_raw, right=self._right_raw)

    def with_both(self) -> LinksTable:
        """
        Add all columns from `left` and `right` with suffixes `_l` and `_r`
        """
        left_columns = [
            _[c].name(c + "_l")
            for c in self.left.columns
            if c + "_l" not in self.columns
        ]
        right_columns = [
            _[c].name(c + "_r")
            for c in self.right.columns
            if c + "_r" not in self.columns
        ]
        x = self
        x = x.with_left(*left_columns)
        x = x.with_right(*right_columns)
        return x

    @property
    def left(self) -> LinkedTable:
        """The left table."""
        from mismo.types._linked_table import LinkedTable

        return LinkedTable(self._left_raw, other=self._right_raw, links=self)

    @property
    def right(self) -> LinkedTable:
        """The right table."""
        from mismo.types._linked_table import LinkedTable

        return LinkedTable(
            self._right_raw, other=self._left_raw, links=self._swap_perspective(self)
        )

    @staticmethod
    def _swap_perspective(links: ibis.Table) -> ibis.Table:
        def _swap_l_and_r(name: str):
            if name.endswith("_l"):
                return name[:-2] + "_r"
            if name.endswith("_r"):
                return name[:-2] + "_l"
            return name

        return links.rename(_swap_l_and_r)

    def cache(self) -> LinksTable:
        """Cache the links table."""
        return LinksTable(
            self.__wrapped__.cache(), left=self._left_raw, right=self._right_raw
        )



from __future__ import annotations

from typing import Literal

import ibis

from mismo.linkage import _linkage
from mismo.linker import _common, _join_linker


class FullLinker(_common.Linker):
    """
    A [Linker][mismo.Linker] that yields all possible pairs (MxN of them).
    """

    def __init__(self, *, task: Literal["dedupe", "link"] | None = None):
        self.task = task

    def __join_condition__(self, left: ibis.Table, right: ibis.Table) -> Literal[True]:
        return True

    def __call__(self, left: ibis.Table, right: ibis.Table) -> _linkage.Linkage:
        return _join_linker.JoinLinker(True, on_slow="ignore", task=self.task)(
            left, right
        )


class EmptyLinker(_common.Linker):
    """A [Linker][mismo.Linker] that yields no pairs."""

    def __init__(self, *, task: Literal["dedupe", "link"] | None = None):
        self.task = task

    def __join_condition__(self, left: ibis.Table, right: ibis.Table) -> Literal[False]:
        return False

    def __call__(self, left: ibis.Table, right: ibis.Table) -> _linkage.Linkage:
        return _join_linker.JoinLinker(False, on_slow="ignore", task=self.task)(
            left, right
        )



"""Utils for cleaning and comparing email addresses."""

from __future__ import annotations

from mismo.lib.phone._core import PhoneMatchLevel as PhoneMatchLevel
from mismo.lib.phone._core import PhonesDimension as PhonesDimension
from mismo.lib.phone._core import clean_phone_number as clean_phone_number
from mismo.lib.phone._core import match_level as match_level



"""Functionalities for text processing and analysis."""

from __future__ import annotations

from mismo.text._features import ngrams as ngrams
from mismo.text._features import tokenize as tokenize
from mismo.text._re_extract import re_extract_struct as re_extract_struct
from mismo.text._similarity import damerau_levenshtein as damerau_levenshtein
from mismo.text._similarity import (
    damerau_levenshtein_ratio as damerau_levenshtein_ratio,
)
from mismo.text._similarity import double_metaphone as double_metaphone
from mismo.text._similarity import jaro_similarity as jaro_similarity
from mismo.text._similarity import jaro_winkler_similarity as jaro_winkler_similarity
from mismo.text._similarity import levenshtein_ratio as levenshtein_ratio
from mismo.text._strings import norm_whitespace as norm_whitespace
from mismo.text._strings import strip_accents as strip_accents



from __future__ import annotations

import ibis
from ibis import _
from ibis.expr import types as ir
import pytest

import mismo
from mismo import KeyLinker
from mismo.tests.util import assert_tables_equal


def blocked_on_letter(t1, t2, **kwargs):
    return ibis.join(
        t1, t2, t1.letter == t2.letter, lname="{name}_l", rname="{name}_r"
    )["record_id_l", "record_id_r"]


def blocked_on_letter_int(t1, t2, **kwargs):
    return ibis.join(
        t1,
        t2,
        (t1.letter == t2.letter) & (t1.int == t2.int),
        lname="{name}_l",
        rname="{name}_r",
    )["record_id_l", "record_id_r"]


def letter_blocked_ids(table_factory):
    """If you block the fixtures t1 and t2 on the letter column,
    these are the record_ids what you should get."""
    return table_factory(
        {
            "record_id_l": [1, 2],
            "record_id_r": [90, 91],
        }
    )


@pytest.mark.parametrize(
    "keys,expected_maker",
    [
        pytest.param(("letter",), blocked_on_letter, id="letter"),
        pytest.param(("letter", "int"), blocked_on_letter_int, id="letter_int"),
        # We can pass a 2-tuple for if we want a different key from each table.
        pytest.param(
            ("letter", ("int", _.int.cast(float))),
            blocked_on_letter_int,
            id="letter_int_lr",
        ),
        pytest.param((_.letter, _.int + 1), blocked_on_letter_int, id="tuple_deferred"),
    ],
)
def test_KeyLinker_keys(t1: ir.Table, t2: ir.Table, keys, expected_maker):
    linkage = KeyLinker(keys)(t1, t2)
    blocked_ids = linkage.links.select("record_id_l", "record_id_r")
    expected = expected_maker(t1, t2)
    assert_tables_equal(blocked_ids, expected)


def test_deduplication(table_factory):
    t = table_factory(
        {
            "record_id": [0, 1, 2, 3, 4],
            "letter": ["a", "b", "a", "b", None],
        }
    )
    expected = table_factory(
        {
            "record_id_l": [0, 1],
            "record_id_r": [2, 3],
        }
    )
    linker = KeyLinker("letter")
    linkage = linker(t, t)
    blocked_ids = linkage.links.select("record_id_l", "record_id_r")
    assert_tables_equal(blocked_ids, expected)


@pytest.mark.xfail
def test_unnest_fails(table_factory, t1: ir.Table, t2: ir.Table):
    # If you do a ibis.join(l, r, _.array.unnest()), that will fail because
    # you can't use unnest in a join condition.
    # But we want to support this case, so test our workaround.
    linkage = KeyLinker(_.array.unnest())(t1, t2)
    blocked_ids = linkage.links.select("record_id_l", "record_id_r")
    expected = table_factory({"record_id_l": [0, 1], "record_id_r": [90, 90]})
    assert_tables_equal(blocked_ids, expected)


@pytest.mark.parametrize(
    "fn",
    [
        pytest.param(
            lambda t: t.inner_join(t.view(), "street").count(),
            id="naive",
        ),
        pytest.param(
            lambda t: mismo.KeyLinker("street").pair_counts(t, t, task="link").n.sum(),
            id="KeyLinker",
        ),
    ],
)
@pytest.mark.parametrize(
    "nrows,exp",
    [
        (1_000, 16_990),
        (10_000, 576_101),
        (100_000, 50_843_827),
        (300_000, 447_872_405),
    ],
)
def test_benchmark_n_pairs(addresses_1M, fn, nrows, exp, benchmark):
    """
    ---------------------------------------------------------------------------------------- benchmark 'test_benchmark_count_pairs': 8 tests ----------------------------------------------------------------------------------------
    Name (time in ms)                                               Min                 Max                Mean             StdDev              Median                IQR            Outliers       OPS            Rounds  Iterations
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    test_benchmark_count_pairs[1000-16990-KeyLinker]            12.2805 (2.39)      14.1857 (2.11)      12.9836 (2.31)      0.3989 (1.19)      12.9002 (2.33)      0.5088 (1.71)         15;2   77.0202 (0.43)         67           1
    test_benchmark_count_pairs[1000-16990-naive]                 5.1490 (1.0)        6.7088 (1.0)        5.6203 (1.0)       0.4985 (1.49)       5.5324 (1.0)       0.7077 (2.37)          2;0  177.9269 (1.0)          11           1
    test_benchmark_count_pairs[10000-576101-KeyLinker]          14.0206 (2.72)      16.7933 (2.50)      15.0547 (2.68)      0.4714 (1.41)      14.9923 (2.71)      0.6295 (2.11)         15;1   66.4246 (0.37)         61           1
    test_benchmark_count_pairs[10000-576101-naive]               8.0191 (1.56)      10.4203 (1.55)       8.7055 (1.55)      0.3354 (1.0)        8.6626 (1.57)      0.2980 (1.0)          24;8  114.8695 (0.65)        101           1
    test_benchmark_count_pairs[100000-50843827-KeyLinker]       31.2830 (6.08)      34.3007 (5.11)      32.7514 (5.83)      0.6897 (2.06)      32.7460 (5.92)      0.6736 (2.26)         10;3   30.5330 (0.17)         31           1
    test_benchmark_count_pairs[100000-50843827-naive]          171.5182 (33.31)    173.2984 (25.83)    172.2893 (30.65)     0.5982 (1.78)     172.2069 (31.13)     0.5607 (1.88)          2;0    5.8042 (0.03)          6           1
    test_benchmark_count_pairs[300000-447872405-KeyLinker]      45.9068 (8.92)      57.3661 (8.55)      52.2799 (9.30)      2.8730 (8.57)      52.7966 (9.54)      3.9854 (13.37)         5;0   19.1278 (0.11)         20           1
    test_benchmark_count_pairs[300000-447872405-naive]         601.2822 (116.78)   652.4741 (97.26)    621.1668 (110.52)   20.6676 (61.63)    612.4148 (110.70)   29.6653 (99.54)         1;0    1.6099 (0.01)          5           1
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    It looks like the naive way starts off only 2x slower for small tables,
    but becomes more than 10x slower for large tables,
    and presumably at much larger tables simply becomes infeasible.
    """  # noqa: E501

    def run():
        return fn(addresses_1M.head(nrows)).execute()

    result = benchmark(run)
    assert result == exp



from __future__ import annotations

import functools
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Literal, cast

import ibis

from mismo import _util
from mismo.types._updates import Updates

if TYPE_CHECKING:
    import altair as alt

    from mismo import _typing


class Diff:
    """A set of insertions, updates, and deletions between two tables.

    This can only semantically represent 1-1 relationships,
    eg a row in the `before` table corresponds to only 0 or 1 row in the `after` table,
    and vice versa. eg "this row changed in these ways between these two tables".

    To represent more general 0-N relationships, use a [Linkage](mismo.Linkage).
    eg many rows in a "dirty" dataset are linked to a single row in a "clean" dataset.
    Say you have a clean database of records.
    You just got a new batch of dirty data that might contain duplicates.
    Each record in the clean database might match multiple records in the dirty data.
    This makes it difficult to use a Diff object, because each clean record
    can't be paired up nicely with a single dirty record.
    A Linkage object is more appropriate in this case.

    This is able to represent a difference between two tables with different schemas,
    eg if a column is added or removed.
    """

    _before: ibis.Table
    _after: ibis.Table
    _insertions: ibis.Table
    _updates: Updates
    _deletions: ibis.Table

    def __new__(*args, **kwargs):
        raise NotImplementedError(
            "Use one of the Diff.from_*() factory methods to create a Diff object."
        )

    @classmethod
    def _new(
        cls,
        before: ibis.Table,
        after: ibis.Table,
        insertions: ibis.Table,
        updates: Updates,
        deletions: ibis.Table,
    ):
        assert isinstance(before, ibis.Table)
        assert isinstance(after, ibis.Table)
        assert isinstance(insertions, ibis.Table)
        assert updates.__class__.__name__ == "Updates"  # isinstance doesn't work??
        assert isinstance(deletions, ibis.Table)

        _check_schemas_equal(before, deletions)
        _check_schemas_equal(before, updates.before())
        _check_schemas_equal(after, insertions)
        _check_schemas_equal(after, updates.after())
        # before and after may have different schemas.

        # only keep updates that actually change something
        any_changed = ibis.or_(*[updates[col].is_changed() for col in updates.columns])
        updates = updates.filter(any_changed)

        obj = super().__new__(cls)
        obj.__init__()
        obj._before = before
        obj._after = after
        obj._insertions = insertions
        obj._updates = updates
        obj._deletions = deletions
        return obj

    @classmethod
    def from_deltas(
        cls,
        *,
        before: ibis.Table,
        insertions: ibis.Table | None = None,
        updates: Updates | None = None,
        deletions: ibis.Table | None = None,
        updates_defaults: None | Any = _util.NOT_SET,
    ) -> _typing.Self:
        """Create from a starting point and a set of transformations.

        Parameters
        ----------
        before
            The table before the changes.
        insertions
            Rows that were in `after` but not in `before`.
            If None, we assume there are no insertions.
        updates
            Rows that were changed between `before` and `after`.
            If None, we assume there are no updates.
        deletions
            Rows that were in `before` but not in `after`.
            If None, we assume there are no deletions.

        Returns
        -------
        Diff
        """
        after = before
        if deletions is not None:
            after = before.difference(deletions, distinct=False)
        else:
            deletions = before.limit(0)

        if updates is not None:
            after = updates.apply_to(after, defaults=updates_defaults)
        else:
            updates = Updates.from_tables(before, after, join_on=False)

        if insertions is not None:
            after = after.union(insertions, distinct=False)
        else:
            insertions = after.limit(0)

        return cls._new(
            before=before,
            after=after,
            insertions=insertions,
            updates=updates,
            deletions=deletions,
        )

    @classmethod
    def from_before_after(
        cls,
        before: ibis.Table,
        after: ibis.Table,
        *,
        join_on: str | Literal[False],
    ) -> _typing.Self:
        """Create from a before and after table."""
        # We have to be careful here.
        # If a row (Bob, 123, 456) is updated to (Bob, 123, 789),
        # Then is this a deletion and insertion, or an update?
        # We have to rely on the join key to determine this,
        # using set-based difference(), intersection(), etc would not work.
        if join_on is False:
            insertions = after
            deletions = before
        else:
            insertions = after.anti_join(before, join_on)
            deletions = before.anti_join(after, join_on)
        updates = Updates.from_tables(before, after, join_on=join_on)
        return cls._new(
            before=before,
            after=after,
            insertions=insertions,
            updates=updates,
            deletions=deletions,
        )

    def cache(self) -> Diff:
        """Return a new Diff with all tables cached."""
        return Diff._new(
            before=self.before().cache(),
            after=self.after().cache(),
            insertions=self.insertions().cache(),
            updates=Updates(self.updates().cache(), check_schemas="lax"),
            deletions=self.deletions().cache(),
        )

    def to_parquets(self, directory: str | Path, /, *, overwrite: bool = False) -> None:
        """Write the tables in the changes to parquet files."""
        directory = Path(directory)
        directory.mkdir(parents=True, exist_ok=True)

        def write(table, name):
            p = directory / f"{name}.parquet"
            if not overwrite and p.exists():
                raise FileExistsError(f"{p} already exists")
            table.to_parquet(p)

        # TODO: this isn't atomic, if one fails, the others will still be written.
        write(self.before(), "before")
        write(self.insertions(), "insertions")
        write(self.deletions(), "deletions")
        write(self.updates(), "updates")

    @classmethod
    def from_parquets(
        cls, directory: str | Path, /, *, backend: ibis.BaseBackend | None = None
    ) -> _typing.Self:
        """Create a Diff by reading parquets from the given directory."""
        if backend is None:
            backend = ibis.get_backend()
        d = Path(directory)
        return cls.from_deltas(
            before=backend.read_parquet(d / "before.parquet"),
            insertions=backend.read_parquet(d / "insertions.parquet"),
            deletions=backend.read_parquet(d / "deletions.parquet"),
            updates=Updates(
                backend.read_parquet(d / "updates.parquet"), check_schemas="lax"
            ),
        )

    def before(self) -> ibis.Table:
        """The table before the changes."""
        return self._before

    def after(self) -> ibis.Table:
        """The table after the changes."""
        return self._after

    def unchanged(self) -> ibis.Table:
        """Rows that were unchanged between `before` and `after`."""
        return self.before().difference(
            self.updates().before(),
            self.deletions(),
        )

    def insertions(self) -> ibis.Table:
        """Rows that were in `after` but not in `before`.

        This has the same schema as `after`."""
        return self._insertions

    def deletions(self) -> ibis.Table:
        """Rows that were in `before` but not in `after`.

        This has the same schema as `before`.
        """
        return self._deletions

    def updates(self) -> Updates:
        """Rows that were changed between `before` and `after`.

        `self.updates().before()` and `self.updates().after()` have the same schemas
        as `before` and `after`, respectively."""
        return self._updates

    def chart(self) -> alt.Chart:
        """Create a chart that shows the flow of rows through the diff.

        This is just a convenience method that calls `self.stats.chart()`.

        ```plaintext
        Rows
        800,000 |                                 ▓▓  Inserted (50,000)
                |                                 ▒▒  Deleted (100,000)
        700,000 |                                 ░░  Updated (200,000)
                |                                 ██  Unchanged (300,000)
        600,000 |      ▒▒▒▒
                |      ▒▒▒▒             ▓▓▓▓
        500,000 |      ░░░░             ░░░░
                |      ░░░░             ░░░░
        400,000 |      ░░░░             ░░░░
                |      ░░░░             ░░░░
        300,000 |      ████             ████
                |      ████             ████
        200,000 |      ████             ████
                |      ████             ████
        100,000 |      ████             ████
                |      ████             ████
              0 | Before (600,000)  After (550,000)
        ```
        """
        return self.stats.chart()

    @property
    def stats(self) -> DiffStats:
        """Statistics about this Diff."""
        return DiffStats(self)

    def __repr__(self):
        return repr(self.stats).replace("DiffStats", "Diff")


def _check_schemas_equal(table1: ibis.Table, table2: ibis.Table):
    pairs1 = set(dict(table1.schema()).items())
    pairs2 = set(dict(table2.schema()).items())
    if pairs1 == pairs2:
        return
    same = pairs1.intersection(pairs2)
    only1 = pairs1 - pairs2
    only2 = pairs2 - pairs1
    raise ValueError(f"Schemas are not equal: {same=}, {only1=}, {only2=}")


class DiffStats:
    """Summary statistics about a Diff, such as number of insertions, deletions, etc."""

    def __init__(self, diff: Diff) -> None:
        """Create from a Diff."""
        self._diff = diff

    @functools.cache
    def n_before(self) -> int:
        """Number of rows in `before`."""
        return int(cast(int, self._diff.before().count().execute()))

    @functools.cache
    def n_unchanged(self) -> int:
        """Number of rows that were unchanged between `before` and `after`."""
        return int(cast(int, self._diff.unchanged().count().execute()))

    @functools.cache
    def n_insertions(self) -> int:
        """Number of rows that were in `after` but not in `before`."""
        return int(cast(int, self._diff.insertions().count().execute()))

    @functools.cache
    def n_deletions(self) -> int:
        """Number of rows that were in `before` but not in `after`."""
        return int(cast(int, self._diff.deletions().count().execute()))

    @functools.cache
    def n_updates(self) -> int:
        """Number of rows that were changed between `before` and `after`."""
        return int(cast(int, self._diff.updates().count().execute()))

    @functools.cache
    def n_after(self) -> int:
        """Number of rows in `after`."""
        return int(cast(int, self._diff.after().count().execute()))

    def __repr__(self):
        return dedent(f"""
        DiffStats(
            before={self.n_before():_},
            after={self.n_after():_},
            unchanged={self.n_unchanged():_},
            updates={self.n_updates():_},
            insertions={self.n_insertions():_},
            deletions={self.n_deletions():_},
        )""")

    def chart(self) -> alt.Chart:
        """Create a chart that shows the flow of rows through the diff.

        ```plaintext
        Rows
        800,000 |                                 ▓▓  Inserted (50,000)
                |                                 ▒▒  Deleted (100,000)
        700,000 |                                 ░░  Updated (200,000)
                |                                 ██  Unchanged (300,000)
        600,000 |      ▒▒▒▒
                |      ▒▒▒▒             ▓▓▓▓
        500,000 |      ░░░░             ░░░░
                |      ░░░░             ░░░░
        400,000 |      ░░░░             ░░░░
                |      ░░░░             ░░░░
        300,000 |      ████             ████
                |      ████             ████
        200,000 |      ████             ████
                |      ████             ████
        100,000 |      ████             ████
                |      ████             ████
              0 | Before (600,000)  After (550,000)
        ```
        """
        import altair as alt

        unchanged_type = f"Unchanged ({self.n_unchanged():,})"
        inserted_type = f"Inserted ({self.n_insertions():,})"
        deleted_type = f"Deleted ({self.n_deletions():,})"
        updated_type = f"Updated ({self.n_updates():,})"

        before = f"Before ({self.n_before():,})"
        after = f"After ({self.n_after():,})"

        def _percent(a, b) -> float:
            return a / b if b else float("nan")

        data = [
            {
                "table": before,
                "type": "Unchanged",
                "type_order": 1,
                "type_detailed": unchanged_type,
                "count": self.n_unchanged(),
                "fraction": _percent(self.n_unchanged(), self.n_before()),
            },
            {
                "table": before,
                "type": "Deleted",
                "type_order": 3,
                "type_detailed": deleted_type,
                "count": self.n_deletions(),
                "fraction": _percent(self.n_deletions(), self.n_before()),
            },
            {
                "table": before,
                "type": "Updated",
                "type_order": 2,
                "type_detailed": updated_type,
                "count": self.n_updates(),
                "fraction": _percent(self.n_updates(), self.n_before()),
            },
            {
                "table": after,
                "type": "Unchanged",
                "type_order": 1,
                "type_detailed": unchanged_type,
                "count": self.n_unchanged(),
                "fraction": _percent(self.n_unchanged(), self.n_after()),
            },
            {
                "table": after,
                "type": "Inserted",
                "type_order": 3,
                "type_detailed": inserted_type,
                "count": self.n_insertions(),
                "fraction": _percent(self.n_insertions(), self.n_after()),
            },
            {
                "table": after,
                "type": "Updated",
                "type_order": 2,
                "type_detailed": updated_type,
                "count": self.n_updates(),
                "fraction": _percent(self.n_updates(), self.n_after()),
            },
        ]
        data = ibis.memtable(data)
        chart = (
            alt.Chart(data)
            .mark_bar()
            .encode(
                alt.X(
                    "table",
                    title=None,
                    sort=[before, after],
                    # axis=alt.Axis(labelAngle=0),
                ),
                alt.Y("count", title="Rows"),
                # Make the unchanged sit on bottom, then updates, deletions insertions
                alt.Order("type_order"),
                alt.Color(
                    "type_detailed",
                    title=None,
                    sort=[
                        inserted_type,
                        deleted_type,
                        updated_type,
                        unchanged_type,
                    ],
                ),
                tooltip=[
                    "type",
                    alt.Tooltip("count", format=","),
                    alt.Tooltip("fraction", format=".2%"),
                ],
            )
        )
        return chart



from __future__ import annotations

from typing import TYPE_CHECKING

import ibis
from ibis import _
from ibis.expr import types as ir

if TYPE_CHECKING:
    import altair as alt
    import ipywidgets


def distribution_chart(vals: ir.Column, *, limit: int | None = None) -> alt.Chart:
    """Make a Altair histogram of values.

    Useful as an exploratory tool to look at what values are present in a column.

    Parameters
    ----------
    vals : ColumnExpr
        The values to plot.
    limit : int, optional
        The maximum number of bars to plot, by default 1000

    Returns
    -------
    alt.Chart
        The histogram.
    """
    import altair as alt

    if limit is None:
        limit = 100
    counts = _make_counts(vals)
    counts = counts.mutate(label=_.value.cast("string"))
    df = counts.to_pandas()
    n_values = df.n.sum()
    n_unique = len(df)
    n_nulls = df[df.value.isnull()].n.sum()
    subset = df.iloc[:limit]

    col = vals.get_name()
    scrubber = alt.selection_interval(encodings=["x"], empty=True)
    x_sort = alt.EncodingSortField("rank", order="ascending")
    overview = (
        alt.Chart(
            subset,
            title=alt.Title(
                text="<Drag to select>",
                dy=30,
                anchor="middle",
                fontSize=12,
                color="gray",
            ),
            width=800,
            height=50,
        )
        .mark_area(interpolate="step-after")
        .encode(
            x=alt.X("label:N", sort=x_sort, axis=None),
            y=alt.Y("n:Q", title=None, axis=None),
        )
        .add_params(scrubber)
    )
    zoomin = (
        alt.Chart(subset, width=800, height=100)
        .mark_bar()
        .encode(
            x=alt.X(
                "label:N",
                sort=x_sort,
                axis=alt.Axis(title=col),
            ),
            y=alt.Y(
                "n:Q",
                title="Count",
                axis=alt.Axis(format="~s"),
            ),
            tooltip=[
                alt.Tooltip("label", title=col),
                alt.Tooltip("n", format="~s", title="Count"),
                alt.Tooltip("pct", format="%", title="Percent"),
            ],
            color=alt.Color("value:N", legend=None),
        )
        .transform_filter(scrubber)
    )
    together = overview & zoomin
    together = together.resolve_scale(color="independent")
    together = together.properties(
        title=alt.Title(
            f"Counts of column '{col}'",
            subtitle=f"showing top {limit:,} values out of {n_values:,} total, {n_unique:,} unique, {n_nulls:,} nulls",  # noqa: E501
            anchor="middle",
            fontSize=14,
        )
    )
    return together


def distribution_dashboard(
    records: ir.Table,
    *,
    column: str | None = None,
    limit: int | None = None,
) -> ipywidgets.VBox:
    """Make an ipywidget dashboard for exploring the distribution of values in a table.

    Parameters
    ----------
    records : Table
        The table to plot.
    column : str, optional
        The initial column to plot. If None, the first column is used.
        You can change this interactively in the returned dashboard.
    limit : int, optional
        The initial maximum number of values to plot, by default 100.
        You can change this interactively in the returned dashboard.

    Returns
    -------
    ipywidgets.VBox
        The dashboard.
    """
    import altair as alt
    import ipywidgets

    if column is None:
        column = records.columns[0]
    limit_max = int(records.count().execute())
    if limit is None:
        limit = min(limit_max, 100)
    column_selector = ipywidgets.Dropdown(
        options=records.columns,
        value=column,
        description="Column:",
        disabled=False,
    )
    limit_selector = ipywidgets.IntSlider(
        min=1, max=limit_max, value=limit, step=1, description="Limit:"
    )

    def get_chart():
        return distribution_chart(
            records[column_selector.value], limit=limit_selector.value
        )

    jupyter_chart = alt.JupyterChart(get_chart())

    def on_change(change):
        jupyter_chart.chart = get_chart()

    column_selector.observe(on_change, names="value")
    limit_selector.observe(on_change, names="value")

    layout = ipywidgets.VBox([column_selector, limit_selector, jupyter_chart])
    return layout


def _make_counts(vals: ir.Column) -> ir.Table:
    vals = vals.name("value")
    counts = vals.value_counts().rename(n="value_count")
    sort_keys = [_.n.desc(), _.value.asc()]
    counts = counts.order_by(sort_keys)
    counts = counts.mutate(rank=ibis.row_number(), pct=_.n / _.n.sum())
    return counts



from __future__ import annotations

import pytest

from mismo._n_naive import n_naive_comparisons


def test_dedupe_with_int():
    assert n_naive_comparisons(0) == 0
    assert n_naive_comparisons(1) == 0
    assert n_naive_comparisons(2) == 1
    assert n_naive_comparisons(3) == 3
    assert n_naive_comparisons(4) == 6
    assert n_naive_comparisons(5) == 10


def test_dedupe_with_sized_list():
    left = [1, 2, 3, 4]
    assert n_naive_comparisons(left) == 4 * 3 // 2


def test_dedupe_with_ibis_table(table_factory):
    t = table_factory({"x": [1, 2, 3]})
    # falls back to count().execute() when len() isn't available
    assert n_naive_comparisons(t) == 3 * 2 // 2


def test_link_with_ints():
    assert n_naive_comparisons(3, 4) == 12


def test_link_with_mixed_sized_and_int():
    left = [1, 2, 3]
    assert n_naive_comparisons(left, 4) == 12
    assert n_naive_comparisons(4, left) == 12


@pytest.mark.usefixtures("backend")
def test_link_with_ibis_tables(table_factory):
    left = table_factory({"x": [1, 2, 3]})
    right = table_factory({"y": [10, 20, 30, 40]})
    assert n_naive_comparisons(left, right) == 12



from __future__ import annotations

import ibis
import pytest

from mismo import text


@pytest.mark.parametrize(
    "inp,exp",
    [
        ("jane's   house", "jane's house"),
        ("Ross' house  ", "Ross' house"),
        ("a", "a"),
        ("", ""),
        (" ", ""),
        (None, None),
        ("bees\tall cook", "bees all cook"),
    ],
)
def test_norm_whitespace(inp, exp):
    inp = ibis.literal(inp, type="string")
    result = text.norm_whitespace(inp).execute()
    assert result == exp



from __future__ import annotations

from ibis.expr import types as ir
import pytest

from mismo.linker import EmptyLinker, FullLinker, Linker
from mismo.tests.util import assert_tables_equal


def test_isinstance():
    assert isinstance(EmptyLinker(), Linker)
    assert isinstance(FullLinker(), Linker)


def test_cant_instantiate():
    with pytest.raises(TypeError):
        Linker()


def test_cross_blocker(table_factory, t1: ir.Table, t2: ir.Table):
    linkage = FullLinker()(t1, t2)
    blocked_ids = linkage.links.select("record_id_l", "record_id_r")
    expected = table_factory(
        {
            "record_id_l": [0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2],
            "record_id_r": [90, 90, 90, 91, 91, 91, 92, 92, 92, 93, 93, 93],
        }
    )
    assert_tables_equal(expected, blocked_ids)


def test_empty_blocker(t1: ir.Table, t2: ir.Table):
    links = EmptyLinker()(t1, t2).links
    assert "record_id_l" in links.columns
    assert "record_id_r" in links.columns
    n = links.count().execute()
    assert n == 0


def test_empty_blocker_self(t1: ir.Table):
    links = EmptyLinker()(t1, t1).links
    assert "record_id_l" in links.columns
    assert "record_id_r" in links.columns
    n = links.count().execute()
    assert n == 0



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, NamedTuple

import ibis
from ibis import _
from ibis.expr import types as ir

from mismo.types._wrapper import TableWrapper

if TYPE_CHECKING:
    import altair as alt


class _HistSpec(NamedTuple):
    n_title: str
    chart_title: str
    chart_subtitle: str


class CountsTable(TableWrapper):
    """A table with at least an Integer column named `n`.

    There will also be variable number of other columns that act as identifiers.

    You won't create this directly, it will be returned to you
    from eg [KeyLinker.key_counts_left][mismo.KeyLinker.key_counts_left],
    [KeyLinker.key_counts_right][mismo.KeyLinker.key_counts_right],
    or [KeyLinker.pair_counts][mismo.KeyLinker.pair_counts].
    """

    n: ir.IntegerColumn
    """The column containing the count."""

    # This MUST be set in subclasses
    _HIST_SPEC: _HistSpec

    @functools.cache
    def n_total(self) -> int:
        """n.sum().fill_null(0), just here for convenience."""
        raw = self.n.sum().execute()
        return int(raw) if raw is not None else 0

    def chart(self) -> alt.Chart:
        return _counts_chart(self, hist_spec=self._HIST_SPEC)


class KeyCountsTable(CountsTable):
    _HIST_SPEC = _HistSpec(
        "Number of Records", "Number of Records by Key", "{n_total:_} Total Records"
    )


class PairCountsTable(CountsTable):
    _HIST_SPEC = _HistSpec(
        "Number of Pairs", "Number of Pairs by Key", "{n_total:_} Total Pairs"
    )


def _counts_chart(counts: CountsTable, *, hist_spec: _HistSpec):
    import altair as alt

    n_total = counts.n_total()
    key_cols = [c for c in counts.columns if c != "n"]
    val = "(" + ibis.literal(", ").join(counts[c].cast(str) for c in key_cols) + ")"
    key_and_n = counts.mutate(val.name("key"), "n")
    key_and_n = key_and_n.filter(_.n > 0)
    frac = key_and_n.n / n_total if n_total > 0 else 0
    key_and_n = key_and_n.mutate(
        frac=frac,
        explanation=(
            "Out of the "
            + f"{n_total:_}, "
            + key_and_n.n.cast(str)
            + " ("
            + (frac * 100).cast(int).cast(str)
            + "%) had the key of "
            + key_and_n.key.cast(str)
        ),
    )
    n_keys_total = key_and_n.count().execute()
    key_title = "(" + ", ".join(key_cols) + ")"
    scrubber_selection = alt.selection_interval(encodings=["x"], empty=True)
    width = 800
    zoomin = (
        alt.Chart(key_and_n, width=width)
        .mark_bar()
        .encode(
            alt.X("key:O", title=key_title, sort="-y"),
            alt.Y(
                "n:Q",
                title=hist_spec.n_title,
                scale=alt.Scale(type="symlog"),
            ),
            tooltip=[
                alt.Tooltip("n:Q", title=hist_spec.n_title, format=","),
                alt.Tooltip("frac:Q", title="Fraction", format=".2%"),
                *[alt.Tooltip(col) for col in key_cols],
                alt.Tooltip("explanation", title="Explanation"),
            ],
        )
        .transform_filter(scrubber_selection)
    )
    scrubber = (
        alt.Chart(
            key_and_n,
            width=width,
            height=50,
            title=alt.Title(
                text="<Drag to select>",
                dy=30,
                anchor="middle",
                fontSize=12,
                color="gray",
            ),
        )
        .mark_area(interpolate="step-after")
        .encode(
            alt.X("key:O", sort="-y", axis=None),
            alt.Y("n:Q", title=None, axis=None),
        )
        .add_params(scrubber_selection)
    )
    together = scrubber & zoomin
    together = together.resolve_scale(color="independent")
    together = together.properties(
        title=alt.Title(
            hist_spec.chart_title,
            subtitle=[
                hist_spec.chart_subtitle.format(n_total=n_total),
                f"{n_keys_total:_} keys total",
            ],
            anchor="middle",
        )
    )
    return together



from __future__ import annotations

from typing import Literal

import ibis

from mismo._resolve import IntoValueResolver, value_resolver
from mismo.linkage import _linkage
from mismo.linker import _common, _join_linker


class UnnestLinker(_common.Linker):
    """A [Linker][mismo.Linker] that unnests a column before linking.

    This is useful if you records with sets of tokens that you want to link on,
    for example:
    - splitting names into words/tokens and linking where any token matches.
    - tags, such as product categories, where you want to link where any tag matches.

    This links where ANY of the unnested values match.

    Examples
    --------
    >>> import ibis
    >>> from ibis import _
    >>> import mismo
    >>> ibis.options.interactive = True
    >>> linkage = mismo.playdata.load_patents()
    >>> t = linkage.left.select("record_id", "name")
    >>> t.head()
    ┏━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ record_id ┃ name                         ┃
    ┡━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
    │ uint32    │ string                       │
    ├───────────┼──────────────────────────────┤
    │      2909 │ * AGILENT TECHNOLOGIES, INC. │
    │      3574 │ * AKZO NOBEL N.V.            │
    │      3575 │ * AKZO NOBEL NV              │
    │      3779 │ * ALCATEL N.V.               │
    │      3780 │ * ALCATEL N.V.               │
    └───────────┴──────────────────────────────┘

    >>> tokens = _.name.upper().split(" ").filter(lambda x: x.length() > 4)
    >>> t.select(tokens.name("tokens"))
    ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ tokens                       ┃
    ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
    │ array<string>                │
    ├──────────────────────────────┤
    │ ['AGILENT', 'TECHNOLOGIES,'] │
    │ ['NOBEL']                    │
    │ ['NOBEL']                    │
    │ ['ALCATEL']                  │
    │ ['ALCATEL']                  │
    │ ['ALCATEL']                  │
    │ ['CANON', 'EUROPA']          │
    │ ['CANON', 'EUROPA']          │
    │ ['CANON', 'EUROPA']          │
    │ []                           │
    │ …                            │
    └──────────────────────────────┘

    Now, block the tables together wherever two records share a token.
    Note that this blocked `* SCHLUMBERGER LIMITED` with `* SCHLUMBERGER TECHNOLOGY BV`.
    because they both share the `SCHLUMBERGER` token.

    >>> linker = mismo.KeyLinker(tokens.unnest())
    >>> linker(t, t).links.filter(_.name_l != _.name_r).order_by(
    ...     "record_id_l", "record_id_r"
    ... ).head()  # doctest: +SKIP
    ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ record_id_l ┃ record_id_r ┃ latitude_l ┃ latitude_r ┃ name_l                                                     ┃ name_r                                                     ┃
    ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
    │ int64       │ int64       │ float64    │ float64    │ string                                                     │ string                                                     │
    ├─────────────┼─────────────┼────────────┼────────────┼────────────────────────────────────────────────────────────┼────────────────────────────────────────────────────────────┤
    │        2909 │    13390969 │        0.0 │      52.35 │ * AGILENT TECHNOLOGIES, INC.                               │ Hitachi Global Storage Technologies, Inc. Netherlands B.V  │
    │        2909 │    13390970 │        0.0 │      52.35 │ * AGILENT TECHNOLOGIES, INC.                               │ Hitachi Global Storage Technologies, Inc. Netherlands B.V. │
    │        2909 │    13391015 │        0.0 │      52.35 │ * AGILENT TECHNOLOGIES, INC.                               │ Hitachi Global Storage Technologies, Netherland B.V.       │
    │        2909 │    13391055 │        0.0 │      52.50 │ * AGILENT TECHNOLOGIES, INC.                               │ Hitachi Global Storage Technologies, Netherlands, B.V.     │
    │        2909 │    13391056 │        0.0 │      52.35 │ * AGILENT TECHNOLOGIES, INC.                               │ Hitachi Global Storage Technologies, Netherlands, B.V.     │
    └─────────────┴─────────────┴────────────┴────────────┴────────────────────────────────────────────────────────────┴────────────────────────────────────────────────────────────┘
    """  # noqa: E501

    def __init__(
        self,
        column: IntoValueResolver,
        *,
        task: Literal["dedupe", "link"] | None = None,
    ):
        self.column_resolver = value_resolver(column)
        self.task = task
        self._linker = _join_linker.JoinLinker(self.column_resolver, task=task)

    def __call__(self, left: ibis.Table, right: ibis.Table) -> _linkage.Linkage:
        resolved_left = self.column_resolver(left)
        resolved_right = self.column_resolver(right)
        left = left.mutate(resolved_left.unnest().name(resolved_left.get_name()))
        right = right.mutate(resolved_right.unnest().name(resolved_right.get_name()))
        return self._linker.__call__(left, right)



from __future__ import annotations

from typing import Any, Callable, Iterable, Literal

import ibis
from ibis import Deferred, _
from ibis.expr import types as ir

from mismo import _resolve, _util
from mismo._counts_table import KeyCountsTable, PairCountsTable
from mismo.linkage import Linkage
from mismo.linker._common import Linker, infer_task
from mismo.types import LinksTable


class KeyLinker(Linker):
    """A [Linker][mismo.Linker] that links records wherever they share a key, eg "emails match."

    This is one of the most basic linking rules, used very often in record linkage.

    Examples
    --------
    >>> import ibis
    >>> from ibis import _
    >>> import mismo
    >>> ibis.options.interactive = True
    >>> linkage = mismo.playdata.load_patents()
    >>> t = linkage.left.select("record_id", "name", "latitude")
    >>> t.head()
    ┏━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━┓
    ┃ record_id ┃ name                         ┃ latitude ┃
    ┡━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━┩
    │ uint32    │ string                       │ float64  │
    ├───────────┼──────────────────────────────┼──────────┤
    │      2909 │ * AGILENT TECHNOLOGIES, INC. │     0.00 │
    │      3574 │ * AKZO NOBEL N.V.            │     0.00 │
    │      3575 │ * AKZO NOBEL NV              │     0.00 │
    │      3779 │ * ALCATEL N.V.               │    52.35 │
    │      3780 │ * ALCATEL N.V.               │    52.35 │
    └───────────┴──────────────────────────────┴──────────┘

    Block the table with itself wherever the names match:

    >>> linker = mismo.KeyLinker("name")
    >>> linker(t, t).links.order_by(
    ...     "record_id_l", "record_id_r"
    ... ).head()  # doctest: +SKIP
    ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━┓
    ┃ record_id_l ┃ record_id_r ┃ latitude_l ┃ latitude_r ┃ name_l         ┃ name_r         ┃
    ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━┩
    │ int64       │ int64       │ float64    │ float64    │ string         │ string         │
    ├─────────────┼─────────────┼────────────┼────────────┼────────────────┼────────────────┤
    │        3779 │        3780 │      52.35 │  52.350000 │ * ALCATEL N.V. │ * ALCATEL N.V. │
    │        3779 │        3782 │      52.35 │   0.000000 │ * ALCATEL N.V. │ * ALCATEL N.V. │
    │        3780 │        3782 │      52.35 │   0.000000 │ * ALCATEL N.V. │ * ALCATEL N.V. │
    │       25388 │     7651559 │       0.00 │  50.966667 │ DSM N.V.       │ DSM N.V.       │
    │       25388 │     7651560 │       0.00 │  52.500000 │ DSM N.V.       │ DSM N.V.       │
    └─────────────┴─────────────┴────────────┴────────────┴────────────────┴────────────────┘

    Arbitrary blocking keys are supported. For example, block the table wherever

    - the first 5 characters of the name in uppercase, are the same
        AND
    - the latitudes, rounded to 1 decimal place, are the same

    >>> linker = mismo.KeyLinker((_["name"][:5].upper(), _.latitude.round(1)))
    >>> linker(t, t).order_by("record_id_l", "record_id_r").head()  # doctest: +SKIP
    ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━┓
    ┃ record_id_l ┃ record_id_r ┃ latitude_l ┃ latitude_r ┃ name_l              ┃ name_r              ┃
    ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━┩
    │ int64       │ int64       │ float64    │ float64    │ string              │ string              │
    ├─────────────┼─────────────┼────────────┼────────────┼─────────────────────┼─────────────────────┤
    │        3574 │        3575 │       0.00 │       0.00 │ * AKZO NOBEL N.V.   │ * AKZO NOBEL NV     │
    │        3779 │        3780 │      52.35 │      52.35 │ * ALCATEL N.V.      │ * ALCATEL N.V.      │
    │       15041 │       15042 │       0.00 │       0.00 │ * CANON EUROPA N.V  │ * CANON EUROPA N.V. │
    │       15041 │       15043 │       0.00 │       0.00 │ * CANON EUROPA N.V  │ * CANON EUROPA NV   │
    │       15042 │       15043 │       0.00 │       0.00 │ * CANON EUROPA N.V. │ * CANON EUROPA NV   │
    └─────────────┴─────────────┴────────────┴────────────┴─────────────────────┴─────────────────────┘
    """  # noqa: E501

    def __init__(
        self,
        keys: str
        | ir.Value
        | Deferred
        | Callable[
            [ibis.Table, ibis.Table],
            tuple[ir.Value | ir.Column, ir.Value | ir.Column],
        ]
        | Iterable[
            str
            | ir.Value
            | Deferred
            | Callable[[ibis.Table], ir.Value | ir.Column | str | Deferred]
            | tuple[
                str
                | Deferred
                | Callable[[ibis.Table], ir.Value | ir.Column | str | Deferred],
                str
                | Deferred
                | Callable[[ibis.Table], ir.Value | ir.Column | str | Deferred],
            ]
            | Callable[
                [ibis.Table, ibis.Table],
                tuple[ir.Value | ir.Column, ir.Value | ir.Column],
            ]
        ],
        *,
        max_pairs: int | None = None,
        task: Literal["dedupe", "lookup", "link"] | None = None,
    ) -> None:
        """Create a KeyLinker.

        Parameters
        ----------
        keys:
            The keys to link on.
            The tables will be linked wherever they share ALL the keys.
            Each key can be any of the following:

            - A string, which is interpreted as the name of a column in both tables.
              eg "price" is equivalent to `left.price == right.price`
            - A Deferred, which is used to reference a column in a table.
              eg `_.price.fill_null(0)` is equivalent to
              `left.price.fill_null(0) == right.price.fill_null(0)`
            - A Callable that takes a table and returns a Column.
            - A 2-tuple of the above, where the first element describes the key in the
              left table and the second element describes the key in the right table.
              eg `("first_name", _.GivenName.upper()")` is equivalent to
              `left.first_name == right.GivenName.upper()`
              This is useful when the keys have different names in the two tables.
            - A callable that takes the left and right tables and returns a tuple
              of columns. Left and right will be joined where the columns are equal.
        max_pairs:
            The maximum number of pairs to generate for each key.
            This is to reduce the impact of very common keys.
            For example, if you are linking people, the name "John Smith" might be
            very common, appearing 5000 times in both left and right.
            This name alone would generate 5000 * 5000 = 25 million pairs,
            which might be too computationally expensive.
            If you set `max_pairs=1000`, then any key that generates more than 1000
            pairs will be ignored.
        task:
            The task to count pairs for.

            - "link": each key results in n_left * n_right pairs
            - "dedupe": each key results in n_left * (n_right - 1) / 2 pairs
               since we will only generate pair (A, B), not also (B, A).
            - None: inferred from the input tables: if `left is right`, then "dedupe",
              otherwise "link".
        """  # noqa: E501
        # TODO: support named keys, eg KeyLinker("age", city=_.city.upper())
        self.resolvers = tuple(_resolve.key_pair_resolvers(keys))
        self.max_pairs = max_pairs
        self.task = task

    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> ir.BooleanValue:
        keys_left, keys_right = self.__keys__(left, right)
        clauses = (kl == kr for kl, kr in zip(keys_left, keys_right, strict=True))
        if self.max_pairs is not None:
            too_common_left, too_common_right = self.too_common_of_records(left, right)
            clauses = [
                *clauses,
                left.record_id.notin(too_common_left.record_id),
                # right.record_id.notin(too_common_right.record_id),
            ]
        return ibis.and_(*clauses)

    def __call__(self, left: ibis.Table, right: ibis.Table) -> Linkage:
        """The linkage between the two tables."""
        # Is this the right place to be calling .view()?
        task = infer_task(task=self.task, left=left, right=right)
        if right is left:
            right = right.view()
        condition = self.__join_condition__(left, right)
        if task == "dedupe":
            condition = condition & (left.record_id < right.record_id)
        return Linkage.from_join_condition(left=left, right=right, condition=condition)

    def too_common_of_records(
        self, left: ibis.Table, right: ibis.Table
    ) -> tuple[ibis.Table, ibis.Table]:
        """The prefilter clause that removes keys that would generate too many pairs."""
        if self.max_pairs is None:
            return left.limit(0), right.limit(0)
        # TODO: this should really use self.pair_counts()
        left_keys, right_keys = self.__keys__(left, right)
        left_key = ibis.struct({k.get_name(): k for k in left_keys}).hash()
        right_key = ibis.struct({k.get_name(): k for k in right_keys}).hash()
        key_name = _util.unique_name()
        # count_name = _util.unique_name()
        left = left.mutate(left_key.name(key_name))
        right = right.mutate(right_key.name(key_name))
        left_counts = left.group_by(key_name).agg(nleft=_.count())
        right_counts = right.group_by(key_name).agg(nright=_.count())
        # If a key is only in one table, it won't be in this joined table,
        # so watch out, we should only test for `IN`, not `NOT IN`
        pair_counts_by_key = ibis.join(
            left_counts,
            right_counts,
            key_name,
        ).select(key_name, npairs=_.nleft * _.nright)
        too_big = pair_counts_by_key.filter(_.npairs > self.max_pairs)
        too_big = too_big.cache()
        left = left.filter(_[key_name].isin(too_big[key_name])).drop(key_name)
        right = right.filter(_[key_name].isin(too_big[key_name])).drop(key_name)
        return left, right

    def __keys__(
        self, left: ibis.Table, right: ibis.Table
    ) -> list[tuple[ir.Column, ir.Column]]:
        lkeys = []
        rkeys = []
        for resolver_l, resolver_r in self.resolvers:
            lkeys.append(resolver_l(left))
            rkeys.append(resolver_r(right))
        return tuple(lkeys), tuple(rkeys)

    def key_counts_left(self, left: ibis.Table, /) -> KeyCountsTable:
        keys = [resolver(left) for resolver, _ in self.resolvers]
        return key_counts(left.select(*keys))

    def key_counts_right(self, right: ibis.Table, /) -> KeyCountsTable:
        keys = [resolver(right) for _, resolver in self.resolvers]
        return key_counts(right.select(*keys))

    def pair_counts(
        self,
        left: ibis.Table,
        right: ibis.Table,
    ) -> PairCountsTable:
        """Count the number of pairs that would be generated by each key.

        If you were to use this blocker to join `left` with `right`,
        how many pairs would be generated for each key?

        This is useful for analyzing the skew of join keys.
        For example, if you are joining on (surname, city),
        there might be only 4 values for (hoessle, tinytown),
        which would lead to a block of 4 * 4 = 16 record pairs.

        On the other hand, there could be 10_000 values for (smith, new york city).
        This would lead to 10_000 * 10_000 = 100_000_000 record pairs,
        which is likely too many for you to be able to compare.

        Parameters
        ----------
        left
            The left table.
        right
            The right table.

        Returns
        -------
        CountsTable
            Will have column(s) for each `key` and a column `n` with the count.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> import mismo
        >>> records = [
        ...     (1, "a", 1),
        ...     (2, "b", 1),
        ...     (3, "b", 1),
        ...     (4, "c", 3),
        ...     (5, "b", 2),
        ...     (6, "c", 3),
        ...     (7, None, 4),
        ...     (8, "c", 3),
        ... ]
        >>> t = ibis.memtable(
        ...     records, schema={"record_id": int, "letter": str, "num": int}
        ... ).cache()
        >>> t
        ┏━━━━━━━━━━━┳━━━━━━━━┳━━━━━━━┓
        ┃ record_id ┃ letter ┃ num   ┃
        ┡━━━━━━━━━━━╇━━━━━━━━╇━━━━━━━┩
        │ int64     │ string │ int64 │
        ├───────────┼────────┼───────┤
        │         1 │ a      │     1 │
        │         2 │ b      │     1 │
        │         3 │ b      │     1 │
        │         4 │ c      │     3 │
        │         5 │ b      │     2 │
        │         6 │ c      │     3 │
        │         7 │ NULL   │     4 │
        │         8 │ c      │     3 │
        └───────────┴────────┴───────┘

        If we joined t with itself using this blocker in a dedupe task,
        we would end up with

        - 3 pairs in the (c, 3) block due to pairs (3, 6), (3, 8), and (6, 8)
        - 1 pairs in the (b, 1) block due to pairs (1, 2)
        - 0 pairs in the (a, 1) block due to record 0 not getting blocked with itself
        - 0 pairs in the (b, 2) block due to record 4 not getting blocked with itself

        >>> linker = mismo.KeyLinker(["letter", "num"], task="dedupe")
        >>> counts = linker.pair_counts(t, t)
        >>> counts.order_by("letter", "num")
        ┏━━━━━━━━┳━━━━━━━┳━━━━━━━┓
        ┃ letter ┃ num   ┃ n     ┃
        ┡━━━━━━━━╇━━━━━━━╇━━━━━━━┩
        │ string │ int64 │ int64 │
        ├────────┼───────┼───────┤
        │ a      │     1 │     0 │
        │ b      │     1 │     1 │
        │ b      │     2 │     0 │
        │ c      │     3 │     3 │
        └────────┴───────┴───────┘

        If we joined t with a copy of itself using this linker in a link task,
        we would end up with

        - 9 pairs in the (c, 3) block due to pairs (3,3), (3, 6), (3, 8), (6, 3), (6, 6), (6, 8), (8, 3), (8, 6), and (8, 8)
        - 4 pairs in the (b, 1) block due to pairs (1, 1), (1, 2), (2, 1), and (2, 2)
        - 1 pairs in the (a, 1) block due to pair (0, 0)
        - 1 pairs in the (b, 2) block due to pair (4, 4)

        >>> linker = mismo.KeyLinker(["letter", "num"], task="link")
        >>> counts = linker.pair_counts(t, t)
        >>> counts.order_by("letter", "num")
        ┏━━━━━━━━┳━━━━━━━┳━━━━━━━┓
        ┃ letter ┃ num   ┃ n     ┃
        ┡━━━━━━━━╇━━━━━━━╇━━━━━━━┩
        │ string │ int64 │ int64 │
        ├────────┼───────┼───────┤
        │ a      │     1 │     1 │
        │ b      │     1 │     4 │
        │ b      │     2 │     1 │
        │ c      │     3 │     9 │
        └────────┴───────┴───────┘

        The returned CountsTable is a subclass of an Ibis Table
        with a special `n_total` method for convenience:

        >>> counts.n_total()
        15
        >>> isinstance(counts, ibis.Table)
        True
        """  # noqa: E501
        too_common_left, too_common_right = self.too_common_of_records(left, right)
        left_filtered = left.filter(_.record_id.notin(too_common_left.record_id))
        right_filtered = right.filter(_.record_id.notin(too_common_right.record_id))
        counts = pair_counts(
            self.resolvers, left_filtered, right_filtered, task=self.task
        )
        if self.max_pairs is not None:
            counts = PairCountsTable(counts.filter(_.n <= self.max_pairs))
        return counts

    def __repr__(self) -> str:
        if self.max_pairs is None:
            max_pairs = "None"
        else:
            max_pairs = f"{self.max_pairs:_}"
        resolvers_str = ", ".join(str(r) for r in self.resolvers)
        return f"{self.__class__.__name__}([{resolvers_str}], max_pairs={max_pairs})"


class KeyLinksTable(LinksTable):
    def __init__(
        self, links: ibis.Table, *, left: ibis.Table, right: ibis.Table, keys, task
    ):
        super().__init__(links, left=left, right=right)
        object.__setattr__(self, "_keys", keys)
        object.__setattr__(self, "_task", task)

    def count(self) -> ibis.ir.IntegerScalar:
        """The number of pairs in the links table."""
        return (
            pair_counts(
                left=self.left_,
                right=self.right_,
                key_pairs=self._keys,
                task=self._task,
            )
            .n.sum()
            .fill_null(0)
            .name("count")
        )


def key_counts(
    key_or_table: ibis.Table | ibis.Value | ibis.Deferred,
    *keys: ibis.Value | ibis.Deferred | Any,
) -> KeyCountsTable:
    t = _util.select(key_or_table, *keys)
    t = t.drop_null(how="any")
    t = t.group_by(t.columns).agg(n=_.count()).order_by(_.n.desc())
    return KeyCountsTable(t)


def pair_counts(
    key_pairs,
    left: ibis.Table,
    right: ibis.Table,
    *,
    task: Literal["dedupe", "link"] | None = None,
) -> PairCountsTable:
    resolvers = _resolve.key_pair_resolvers(key_pairs)
    keys_left = [resolver(left) for resolver, _ in resolvers]
    keys_right = [resolver(right) for _, resolver in resolvers]
    kcl = key_counts(left.select(*keys_left))
    kcr = key_counts(right.select(*keys_right))
    task = infer_task(task=task, left=left, right=right)
    if task == "dedupe":
        by_key = kcl.mutate(n=(_.n * (_.n - 1) / 2).cast(int))
    else:
        if kcl.equals(kcr):
            kcr = kcr.view()
        condition = ibis.and_(
            *(
                kcl[a] == kcr[b]
                for a, b in zip(kcl.columns, kcr.columns, strict=True)
                if a != "n" and b != "n"
            )
        )
        by_key = (
            ibis.join(kcl, kcr, condition).mutate(n=_.n * _.n_right).drop("n_right")
        )
    by_key = by_key.order_by(_.n.desc())
    return PairCountsTable(by_key)



from __future__ import annotations

from mismo.linkage._linkage import Linkage as Linkage
from mismo.linkage._linkage import filter_links as filter_links
from mismo.linkage._sample import sample_all_links as sample_all_links



from __future__ import annotations

import ibis
from ibis.expr import types as ir

from mismo import _structs, _util
from mismo.compare import MatchLevel
from mismo.lib.name._nicknames import are_aliases
from mismo.text import damerau_levenshtein


def are_match_with_nicknames(
    left: ir.StructValue, right: ir.StructValue
) -> ir.BooleanValue | bool:
    """The given names match via nickname or alias, and the surname names match."""
    return ibis.and_(
        are_aliases(left.given, right.given),
        left.surname == right.surname,
    )


def initials_equal(
    left: ir.StringValue, right: ir.StringValue
) -> ir.BooleanValue | bool:
    """The first letter matches, and at least one is a single letter."""
    return ibis.and_(
        left[0] == right[0],
        ibis.or_(right.length() == 1, left.length() == 1),
    )


def equal_forgiving_typo(
    left: ir.StringValue, right: ir.StringValue
) -> ir.BooleanValue | bool:
    edit_distance = damerau_levenshtein(left, right)
    return ibis.or_(
        edit_distance <= 1,
        ibis.and_(edit_distance <= 2, left.length() > 5),
        _substring_match(left, right),
    )


def _substring_match(
    left: ir.StringValue, right: ir.StringValue, *, min_len: int = 3
) -> ir.BooleanValue | bool:
    """The shorter string is a substring of the longer string, and at least min_len."""
    return ibis.or_(
        ibis.and_(left.contains(right), right.length() >= min_len),
        ibis.and_(right.contains(left), left.length() >= min_len),
    )


class NameMatchLevel(MatchLevel):
    """How closely two names match."""

    NULL = 0
    """At least one given or surname is NULL from either side."""
    EXACT = 1
    """The names are exactly the same."""
    GIVEN_SURNAME = 2
    """The given and surnames both match."""
    NICKNAMES = 3
    """The given names match with nicknames, and the surnames match."""
    INITIALS = 4
    """The first letter of the given name matches, and the surnames match."""
    TYPO = 5
    """The given names are the same (forgiving typos), and the surnames match."""
    ELSE = 6
    """None of the above."""


def _get_level(le: ir.StructValue, ri: ir.StructValue) -> ir.IntegerValue:
    # assumes the inputs have already been featurized/normalized
    # with _clean.normalize_name()
    return _util.cases(
        (
            ibis.or_(
                _structs.struct_isnull(le, how="any", fields=["given", "surname"]),
                _structs.struct_isnull(ri, how="any", fields=["given", "surname"]),
            ),
            NameMatchLevel.NULL.as_integer(),
        ),
        (_structs.struct_equal(le, ri), NameMatchLevel.EXACT.as_integer()),
        (
            _structs.struct_equal(le, ri, fields=["given", "surname"]),
            NameMatchLevel.GIVEN_SURNAME.as_integer(),
        ),
        (
            are_match_with_nicknames(le, ri),
            NameMatchLevel.NICKNAMES.as_integer(),
        ),
        (
            ibis.and_(
                equal_forgiving_typo(le.given, ri.given),
                le.surname == ri.surname,
            ),
            NameMatchLevel.TYPO.as_integer(),
        ),
        (
            ibis.and_(
                initials_equal(le.given, ri.given),
                le.surname == ri.surname,
            ),
            NameMatchLevel.INITIALS.as_integer(),
        ),
        else_=(NameMatchLevel.ELSE.as_integer()),
    )


class NameComparer:
    """Compare names. Assumes the names have already been normalized/featurized."""

    Levels = NameMatchLevel
    """The levels of name comparison."""

    def __init__(
        self,
        left_column: str = "name_l",
        right_column: str = "name_r",
        *,
        result_column: str | None = None,
    ):
        if result_column is None:
            result_column = f"{left_column}_vs_{right_column}"
        self.left_column = left_column
        self.right_column = right_column
        self.name = result_column

    def __call__(
        self,
        table: ir.Table | None = None,
        *,
        left: ir.StructValue | None = None,
        right: ir.StructValue | None = None,
    ) -> ir.IntegerValue:
        """Compare two names.

        Parameters
        ----------
        left :
            The left name.
        right :
            The right name.

        Returns
        -------
        t :
            The comparison result.
        """
        if table is None and (left is None or right is None):
            raise ValueError("Must provide either table or left and right")
        left = left or table[self.left_column]
        right = right or table[self.right_column]
        return _get_level(left, right).name(self.name)



from __future__ import annotations

import ibis
import pandas._testing as tm
import pytest

from mismo.lib import name


@pytest.mark.parametrize(
    ("name1", "name2", "expected"),
    [
        ("alex", "alex", True),
        (" ALEX ", "alex", True),
        ("al", "alex", True),
        ("alex", "al", True),
        ("alex", "alexander", True),
        ("al", "alexander", True),
        ("alexa", "alexander", False),
        ("zach", "zack", True),
        ("betsy", "jon", False),
        ("mary", "mary", True),
        ("bob", "bob", True),
        ("robert", "bob", True),
        ("bob", "robert", True),
        ("bob", "  ROBERT", True),
        ("robert", "mary", False),
        ("robert", "robert", True),
        ("robert", "roberta", False),
        (ibis.literal("robert"), "bob", True),
        (ibis.literal("bob"), "bob", True),
    ],
)
def test_are_aliases(name1, name2, expected):
    result = name.are_aliases(name1, name2)
    if isinstance(result, ibis.Expr):
        result = result.execute()
    assert expected == result


def test_are_aliases_column():
    t = ibis.memtable(
        {
            "a": ["robert", "bob", "robert", "rob"],
            "b": ["mary", "robert", "rob", "rob"],
            "expected": [False, True, True, True],
        }
    )
    res = name.are_aliases(t.a, t.b)
    print(id(res._find_backend(use_default=True)))
    result = t.mutate(aliases=res).execute()
    tm.assert_series_equal(result.expected, result.aliases, check_names=False)



from __future__ import annotations

from collections.abc import Iterable
from itertools import combinations
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import altair as alt
    import pandas as pd


def combos(set_names: Iterable[str]) -> frozenset[frozenset[str]]:
    """
    Computes all non-empty subsets of a given iterable of set names.

    Args:
        set_names: An iterable of strings, where each string is a set name.

    Returns:
        A frozenset of frozensets, where each set represents a unique combination
        of the input set names. For example, given ['A', 'B'], it will
        return {{'A'}, {'B'}, {'A', 'B'}}.
    """
    s = list(set_names)
    result = []
    for r in range(1, len(s) + 1):
        result.extend(frozenset(combo) for combo in combinations(s, r))
    return frozenset(result)


assert combos(["A", "B"]) == frozenset(
    {frozenset({"A"}), frozenset({"B"}), frozenset({"A", "B"})}
)
assert combos(["A", "B", "C"]) == frozenset(
    {
        frozenset({"A"}),
        frozenset({"B"}),
        frozenset({"C"}),
        frozenset({"A", "B"}),
        frozenset({"A", "C"}),
        frozenset({"B", "C"}),
        frozenset({"A", "B", "C"}),
    }
)


def upset_chart(data: Any) -> alt.Chart:
    """Generate an UpSet plot.

    Parameters
    ----------
    data
        The data to plot. A Pandas DataFrame or anything that supports the
        __dataframe__ protocol, with each row representing a single
        intersection between sets.
        There should be columns:

        - A column containing the size of the intersection called "intersection_size"
        - A column for each set, with a boolean value indicating whether
          the intersection is in that set.
        - There should be no other columns.

    Returns
    -------
    Chart
        An Altair chart.
    """
    import altair as alt

    df = _to_df(data)
    longer = _pivot_longer(df)
    sets = (
        longer[longer.is_intersect]
        .groupby("set")["intersection_size"]
        .sum()
        .sort_values(ascending=False)
        .index.tolist()
    )
    base = alt.Chart(longer)
    intersection_x = alt.X(
        "intersection_id:O",
        axis=None,
        sort=alt.EncodingSortField("intersection_size"),
    )
    set_y = alt.Y(
        "set:O",
        sort=sets,
        axis=alt.Axis(title="Blocking Rule", labelLimit=500),
    )
    p_intersection = _intersection_plot(base, sets, intersection_x)
    p_matrix = _matrix_plot(base, sets, intersection_x, set_y)
    p_sets = _set_plot(base, set_y)
    return alt.vconcat(
        p_intersection,
        alt.hconcat(
            p_matrix,
            p_sets,
            spacing=0,
        ),
        spacing=0,
    )


def _to_df(data) -> pd.DataFrame:
    import pandas as pd

    if isinstance(data, pd.DataFrame):
        return data
    try:
        interchange_object = data.__dataframe__()
    except AttributeError:
        pass
    else:
        return pd.api.interchange.from_dataframe(interchange_object)
    return pd.DataFrame(data)


def _pivot_longer(df: pd.DataFrame) -> pd.DataFrame:
    """Resulting DF has one row for each cell in the matrix.

        intersection_id  intersection_size  intersection_degree                set  is_intersect  set_order
    0                 0                 45                    2       Name First 3          True          0
    1                 1             111258                    2       Name First 3          True          0
    2                 2                 38                    3       Name First 3          True          0
    3                 3               1225                    3       Name First 3          True          0
    4                 4                  2                    2       Name First 3         False          0
    5                 5                906                    3       Name First 3          True          0
    6                 6                  6                    1       Name First 3         False          0
    7                 7               2470                    2       Name First 3          True          0
    8                 8               6340                    1       Name First 3         False          0
    9                 9                646                    2       Name First 3         False          0
    10               10                  2                    3       Name First 3         False          0
    11               11               1495                    4       Name First 3          True          0
    12               12             204224                    1       Name First 3         False          0
    13               13             303104                    1       Name First 3          True          0
    14                0                 45                    2  Coordinates Close         False          1
    15                1             111258                    2  Coordinates Close          True          1
    16                2                 38                    3  Coordinates Close          True          1
    17                3               1225                    3  Coordinates Close          True          1
    18                4                  2                    2  Coordinates Close         False          1
    19                5                906                    3  Coordinates Close         False          1
    20                6                  6                    1  Coordinates Close         False          1
    21                7               2470                    2  Coordinates Close         False          1
    22                8               6340                    1  Coordinates Close         False          1
    23                9                646                    2  Coordinates Close          True          1
    24               10                  2                    3  Coordinates Close          True          1
    25               11               1495                    4  Coordinates Close          True          1
    26               12             204224                    1  Coordinates Close          True          1
    27               13             303104                    1  Coordinates Close         False          1
    28                0                 45                    2    Coauthors Exact          True          2
    29                1             111258                    2    Coauthors Exact         False          2
    30                2                 38                    3    Coauthors Exact          True          2
    31                3               1225                    3    Coauthors Exact         False          2
    32                4                  2                    2    Coauthors Exact          True          2
    33                5                906                    3    Coauthors Exact          True          2
    34                6                  6                    1    Coauthors Exact          True          2
    35                7               2470                    2    Coauthors Exact         False          2
    36                8               6340                    1    Coauthors Exact         False          2
    37                9                646                    2    Coauthors Exact         False          2
    38               10                  2                    3    Coauthors Exact          True          2
    39               11               1495                    4    Coauthors Exact          True          2
    40               12             204224                    1    Coauthors Exact         False          2
    41               13             303104                    1    Coauthors Exact         False          2
    42                0                 45                    2      Classes Exact         False          3
    43                1             111258                    2      Classes Exact         False          3
    44                2                 38                    3      Classes Exact         False          3
    45                3               1225                    3      Classes Exact          True          3
    46                4                  2                    2      Classes Exact          True          3
    47                5                906                    3      Classes Exact          True          3
    48                6                  6                    1      Classes Exact         False          3
    49                7               2470                    2      Classes Exact          True          3
    50                8               6340                    1      Classes Exact          True          3
    51                9                646                    2      Classes Exact          True          3
    52               10                  2                    3      Classes Exact          True          3
    53               11               1495                    4      Classes Exact          True          3
    54               12             204224                    1      Classes Exact         False          3
    55               13             303104                    1      Classes Exact         False          3
    """  # noqa: E501
    import pandas as pd

    sets = [c for c in df.columns if c != "intersection_size"]
    df["intersection_id"] = range(len(df))
    df["intersection_degree"] = df[sets].sum(axis=1)
    longer = pd.melt(
        df, id_vars=["intersection_id", "intersection_size", "intersection_degree"]
    )
    longer = longer.rename(columns={"variable": "set", "value": "is_intersect"})
    set_mapping = {s: i for i, s in enumerate(sets)}
    longer["set_order"] = longer["set"].map(set_mapping)
    return longer


def _intersection_plot(base: alt.Chart, sets: Iterable[str], x) -> alt.Chart:
    import altair as alt

    sets = list(sets)
    intersection_base = base.transform_filter(alt.datum.set == sets[0]).encode(
        x=x,
        y=alt.Y("intersection_size:Q", title="Number of Pairs"),
    )
    bars = intersection_base.mark_bar(color="black")
    text = intersection_base.mark_text(
        dx=5,
        angle=270,
        align="left",
    ).encode(text=alt.Text("intersection_size:Q", format=","))
    layered = bars + text
    return layered


def _matrix_plot(base: alt.Chart, sets: Iterable[str], x, y) -> alt.Chart:
    import altair as alt

    sets = list(sets)
    matrix_circle_bg = base.mark_circle(size=100, color="lightgray", opacity=1).encode(
        x=x,
        y=y,
        tooltip=["set:N"],
    )
    matrix_circle = matrix_circle_bg.mark_circle(
        size=150, color="black", opacity=1
    ).transform_filter(alt.datum.is_intersect)
    layered = alt.layer(matrix_circle_bg, matrix_circle)
    return layered


def _set_plot(base: alt.Chart, y) -> alt.Chart:
    import altair as alt

    set_base = base.transform_filter(alt.datum.is_intersect).encode(
        x=alt.X("sum(intersection_size):Q", title="Number of Pairs"),
        y=y.axis(None),
        tooltip=["set:N"],
    )
    set_bars = set_base.mark_bar().encode(color=alt.Color("set:N", legend=None))
    set_text = set_base.mark_text(dx=5, align="left").encode(
        text=alt.Text("sum(intersection_size):Q", format=",")
    )
    layered = set_bars + set_text
    layered = layered.properties(width=150)
    return layered



from __future__ import annotations

from collections.abc import Iterable, Mapping

import ibis
from ibis import _
import solara

from mismo.cluster import degree


@solara.component
def degree_dashboard(
    tables: ibis.Table | Iterable[ibis.Table] | Mapping[str, ibis.Table],
    links: ibis.Table,
) -> None:
    with_degree = degree(tables, links)

    def degree_hist(links_by_record):
        return (
            links_by_record.group_by(_.degree.name("Degree"))
            .agg(_.count().name("Number of Records"))
            .order_by("Degree")
        )

    solara.Markdown(
        """
        ### Distribution of Degrees

        The "degree" of a record is the number of other records it is linked to. Use this to find how many records
        
        - were not linked
        - were linked to just one other record
        - were linked to many other records
        """,  # noqa: E501
        style="align: center",
    )
    with solara.Columns():
        for name, table in with_degree.items():
            with solara.Column():
                solara.Markdown(f"### {name}", style="text-align: center")
                solara.DataFrame(degree_hist(table).execute())



from __future__ import annotations

import ibis
from ibis import _
import pytest

from mismo import Linkage, LinkedTable
from mismo.tests.util import assert_tables_equal


@pytest.fixture
def linkage() -> Linkage:
    left = ibis.memtable({"record_id": [4, 5, 6]})
    right = ibis.memtable({"record_id": [7, 8, 9], "extra": ["a", "b", "c"]})
    links = ibis.memtable(
        {"record_id_l": [4, 4, 5], "record_id_r": [7, 8, 9], "extra": [1, 2, 3]}
    )
    return Linkage(left=left, right=right, links=links)


def test_Linkage_init():
    left = ibis.memtable({"record_id": [4, 5, 6]})
    right = ibis.memtable({"record_id": [7, 8, 9]})
    links = ibis.memtable(
        {"record_id_l": [4, 4, 5], "record_id_r": [7, 8, 9], "extra": [1, 2, 3]}
    )
    # no error on extra column
    Linkage(left=left, right=right, links=links)


def test_LinkedTable_with_n_links(linkage: Linkage):
    actual = linkage.left.with_n_links()
    expected = ibis.memtable(
        {
            "record_id": [4, 5, 6],
            "n_links": [2, 1, 0],
        }
    )
    assert_tables_equal(expected, actual)

    actual = linkage.right.with_n_links(name="foo")
    expected = ibis.memtable(
        {
            "record_id": [7, 8, 9],
            "extra": ["a", "b", "c"],
            "foo": [1, 1, 1],
        }
    )
    assert_tables_equal(expected, actual)


def test_LinkedTable_link_counts(linkage: Linkage):
    actual = linkage.left.link_counts()
    # ┏━━━━━━━━━┳━━━━━━━━━━━┓
    # ┃ n_links ┃ n_records ┃
    # ┡━━━━━━━━━╇━━━━━━━━━━━┩
    # │ int64   │ int64     │
    # ├─────────┼───────────┤
    # │       2 │         1 │  # there was 1 record in tl that matched 2 in tr
    # │       0 │         1 │  # there was 1 record in tl that didn't match any in tr
    # │       1 │         1 │  # there was 1 record in tl that matched 1 in tr
    # └─────────┴───────────┘
    expected = ibis.memtable(
        {
            "n_records": [1, 1, 1],
            "n_links": [0, 1, 2],
        }
    )
    assert_tables_equal(expected, actual, column_order="ignore")

    actual = linkage.right.link_counts()
    # ┏━━━━━━━━━┳━━━━━━━━━━━┓
    # ┃ n_links ┃ n_records ┃
    # ┡━━━━━━━━━╇━━━━━━━━━━━┩
    # │ int64   │ int64     │
    # ├─────────┼───────────┤
    # │       1 │         3 │  # there were 3 records in tr that matched 1 in tl
    # └─────────┴───────────┘
    expected = ibis.memtable(
        {
            "n_records": [3],
            "n_links": [1],
        }
    )
    assert_tables_equal(expected, actual, column_order="ignore")


def test_Linkage_link_counts_chart(linkage: Linkage):
    # just a smoketest that it doesn't crash
    linkage.link_counts_chart()


def test_Linkage_empty_link_counts_chart(linkage: Linkage):
    empty_linkage = Linkage(
        left=linkage.left, right=linkage.right, links=linkage.links.limit(0)
    )
    empty_linkage.link_counts_chart()


def test_Linkage_with_linked_values(linkage: Linkage, table_factory):
    result = linkage.left.with_linked_values(
        "extra",
        x=_.record_id.max(),
        y=(_.record_id + 1),
    )
    assert isinstance(result, LinkedTable)

    expected = table_factory(
        [
            (4, ["a", "b"], 8, [8, 9]),
            (5, ["c"], 9, [10]),
            (6, None, None, None),
        ],
        schema={
            "record_id": "int64",
            "extra": "array<string>",
            "x": "int64",
            "y": "array<int64>",
        },
    )
    assert_tables_equal(expected, result)



from __future__ import annotations

import ibis
from ibis import _
import numpy as np
import pytest

from mismo.lib.geo import CoordinateLinker, distance_km


@pytest.mark.parametrize(
    "lat1, lon1, lat2, lon2, expected",
    [
        (0, 0, 0, 0, 0),
        (None, 0, 0, 0, np.nan),
        (0, 0, 0, 180, 20015.086796020572),
        (0, 0, 0, 90, 10007.543398010286),
    ],
)
def test_distance_km(lat1, lon1, lat2, lon2, expected):
    lat1 = ibis.literal(lat1, type="double")
    lon1 = ibis.literal(lon1, type="double")
    lat2 = ibis.literal(lat2, type="double")
    lon2 = ibis.literal(lon2, type="double")
    result = distance_km(lat1=lat1, lon1=lon1, lat2=lat2, lon2=lon2).execute()
    assert result == pytest.approx(expected) or np.isnan(result) and np.isnan(expected)


@pytest.mark.parametrize(
    "coord1,coord2,km,expected",
    [
        ((0, 0), (0, 0), 1, True),
        pytest.param(
            (61.1547800, -150.0677490),
            (61.1582056, -150.0584552),
            1,
            True,
            id=".6km-anchorage-1km",
        ),
        pytest.param(
            (61.1547800, -150.0677490),
            (61.1582056, -150.0584552),
            0.1,
            False,
            id=".6km-anchorage-100m",
        ),
        # tests for https://github.com/NickCrews/mismo/issues/74
        pytest.param((0, 0), (0, 0.5), 1, False, id="50km-equator-1km"),
        pytest.param((0, 0), (0, 0.5), 100, True, id="50km-equator-100km"),
    ],
)
@pytest.mark.parametrize(
    "kwargs",
    [
        pytest.param(
            {"right_resolver": {"lat": _.coord.lat, "lon": _.coord.lon}},
            id="dict_of_deferred",
        ),
        pytest.param(
            {"right_resolver": lambda t: {"lat": t.coord.lat, "lon": t.coord.lon}},
            id="callable",
        ),
        pytest.param(
            {"right_resolver": "coord"},
            id="coord_name",
        ),
    ],
)
def test_coordinate_blocker(table_factory, coord1, coord2, km, expected, kwargs):
    lat1, lon1 = coord1
    lat2, lon2 = coord2
    t1 = table_factory(
        {
            "lat": [lat1],
            "lon": [lon1],
            "record_id": [42],
        }
    )
    t2 = table_factory(
        {
            "coord": [{"lat": lat2, "lon": lon2}],
            "record_id": [53],
        }
    )
    blocker = CoordinateLinker(distance_km=km, **kwargs)
    linkage = blocker(t1, t2)
    n_blocked = linkage.links.count().execute()
    if expected:
        assert n_blocked == 1
    else:
        assert n_blocked == 0



from __future__ import annotations

from ibis.expr import types as ir
import pytest

from mismo import cluster


@pytest.fixture
def labels_true(table_factory) -> ir.Table:
    return table_factory(
        {
            "record_id": [0, 1, 2, 3],
            "label": [0, 0, 0, 1],
        }
    )


@pytest.fixture
def labels_pred(table_factory) -> ir.Table:
    return table_factory(
        {
            "record_id": [0, 1, 2, 3],
            "label": [0, 0, 1, 1],
        }
    )


@pytest.mark.parametrize(
    "func",
    [
        cluster.rand_score,
        cluster.adjusted_rand_score,
        cluster.fowlkes_mallows_score,
        cluster.homogeneity_score,
        cluster.completeness_score,
        cluster.v_measure_score,
        cluster.mutual_info_score,
        cluster.normalized_mutual_info_score,
    ],
)
def test_scalar_metrics(func, labels_true: ir.Table, labels_pred: ir.Table):
    # We assume sklearns implementation is correct
    result = func(labels_true, labels_pred)
    assert result >= 0.0
    assert result <= 1.0
    assert func.__doc__ is not None


def test_homogeneity_completeness_v_measure(
    labels_true: ir.Table, labels_pred: ir.Table
):
    # We assume sklearns implementation is correct
    homogeneity, completeness, v_measure = cluster.homogeneity_completeness_v_measure(
        labels_true, labels_pred
    )
    assert homogeneity >= 0.0
    assert homogeneity <= 1.0
    assert completeness >= 0.0
    assert completeness <= 1.0
    assert v_measure >= 0.0
    assert v_measure <= 1.0



from __future__ import annotations

from ibis import _
import pytest

from mismo.tests.util import assert_tables_equal
from mismo.types import Diff, Updates


@pytest.fixture
def before(table_factory):
    return table_factory(
        {
            "id": [1, 2, 3, 4],
            "name": ["Alice", None, "Charlie", "David"],
            "age": [10, 20, 30, 40],
        }
    )


@pytest.fixture
def updates_after(table_factory):
    return table_factory(
        {
            "id": [1, 2],
            "name": ["Alice", "Bob"],
            "age": [10, 20],
        }
    )


@pytest.fixture
def updates(before, updates_after):
    return Updates.from_tables(before, updates_after, join_on="id")


@pytest.fixture
def insertions(table_factory):
    return table_factory(
        {
            "id": [5],
            "name": ["Eve"],
            "age": [50],
        }
    )


@pytest.fixture
def deletions(table_factory):
    return table_factory(
        {
            "id": [3],
            "name": ["Charlie"],
            "age": [30],
        }
    )


@pytest.fixture
def diff(before, updates, insertions, deletions):
    return Diff.from_deltas(
        before=before, updates=updates, insertions=insertions, deletions=deletions
    )


@pytest.fixture
def after(table_factory):
    return table_factory(
        {
            "id": [1, 2, 4, 5],
            "name": ["Alice", "Bob", "David", "Eve"],
            "age": [10, 20, 40, 50],
        }
    )


def test_direct_init_raises():
    """You can only create from the factory methods."""
    with pytest.raises(NotImplementedError):
        Diff(before=None, after=None, insertions=None, deletions=None, updates=None)


def test_from_before_after(after, before, updates, insertions, deletions):
    diff = Diff.from_before_after(before, after, join_on="id")
    expected_updates = updates.filter(_.id.before == 2)  # the id==1 row is unchanged
    assert_tables_equal(before, diff.before(), order_by="id")
    assert_tables_equal(after, diff.after(), order_by="id")
    assert_tables_equal(insertions, diff.insertions(), order_by="id")
    assert_tables_equal(deletions, diff.deletions(), order_by="id")
    assert_tables_equal(expected_updates, diff.updates(), order_by="id")


def test_from_before_after_no_join(after, before):
    diff = Diff.from_before_after(before, after, join_on=False)
    assert diff.updates().count().execute() == 0
    assert_tables_equal(before, diff.before(), order_by="id")
    assert_tables_equal(after, diff.after(), order_by="id")
    assert_tables_equal(after, diff.insertions(), order_by="id")
    assert_tables_equal(before, diff.deletions(), order_by="id")


def test_from_deltas(after, before, updates, insertions, deletions):
    diff = Diff.from_deltas(
        before=before, updates=updates, insertions=insertions, deletions=deletions
    )
    expected_updates = updates.filter(_.id.before == 2)  # the id==1 row is unchanged
    assert_tables_equal(before, diff.before(), order_by="id")
    assert_tables_equal(after, diff.after(), order_by="id")
    assert_tables_equal(insertions, diff.insertions(), order_by="id")
    assert_tables_equal(deletions, diff.deletions(), order_by="id")
    assert_tables_equal(expected_updates, diff.updates(), order_by="id")


def test_from_deltas_noop(before):
    """
    With no deltas, the before and after are the same, and the other tables are empty.
    """
    diff = Diff.from_deltas(before=before)
    assert_tables_equal(before, diff.before(), order_by="id")
    assert_tables_equal(before, diff.after(), order_by="id")
    assert_tables_equal(before.limit(0), diff.insertions(), order_by="id")
    assert_tables_equal(before.limit(0), diff.deletions(), order_by="id")
    assert_tables_equal(
        Updates.from_tables(before, before, join_on="id").limit(0),
        diff.updates(),
        order_by="id",
    )


def test_parquet_roundtrip(diff, tmp_path):
    diff.to_parquets(tmp_path)
    reloaded = Diff.from_parquets(tmp_path)
    assert_tables_equal(diff.before(), reloaded.before(), order_by="id")
    assert_tables_equal(diff.after(), reloaded.after(), order_by="id")
    assert_tables_equal(diff.insertions(), reloaded.insertions(), order_by="id")
    assert_tables_equal(diff.deletions(), reloaded.deletions(), order_by="id")
    assert_tables_equal(diff.updates(), reloaded.updates(), order_by="id")



from __future__ import annotations

import dataclasses
import os
from pathlib import Path
from typing import Any, Callable, Iterable, Protocol

import ibis
from ibis import _
from ibis.expr import datatypes as dt
from ibis.expr import types as ir
import pytest

try:
    # Workaround for https://github.com/explosion/thinc/pull/965
    import thinc.util

    _original_fix_random_seed = thinc.util.fix_random_seed

    def _patched_fix_random_seed(seed: int = 0) -> None:
        seed = seed % (2**32)
        _original_fix_random_seed(seed)

    thinc.util.fix_random_seed = _patched_fix_random_seed
except ImportError:
    pass

# we want to have pytest assert introspection in the helpers
pytest.register_assert_rewrite("mismo.tests.util")

DATA_DIR = Path(__file__).parent.parent / "data"


@pytest.fixture
def backend() -> ibis.BaseBackend:
    return ibis.duckdb.connect()


_count = 0


@pytest.fixture
def table_factory(backend: ibis.BaseBackend) -> Callable[..., ir.Table]:
    def factory(data, schema=None, columns=None, **kwargs):
        global _count
        name = f"__mismo_test{_count}"
        _count += 1
        mt = ibis.memtable(data, schema=schema, columns=columns)
        result = backend.create_table(name, mt, **kwargs)
        return result

    return factory


class ColumnFactory(Protocol):
    def __call__(
        column_data: Iterable[Any], type: str | dt.DataType | None = None
    ) -> ir.Column: ...


@pytest.fixture
def column_factory(table_factory) -> ColumnFactory:
    def func(column_data: Iterable[Any], type: str | dt.DataType | None = None):
        if type is not None:
            table = table_factory({"column": column_data}, schema={"column": type})
        else:
            table = table_factory({"column": column_data})
        return table.column

    return func


@pytest.fixture
def t1(table_factory) -> ir.Table:
    return table_factory(
        {
            "record_id": [0, 1, 2],
            "int": [1, 2, 3],
            "letter": ["a", "b", "c"],
            "array": [["a", "b"], ["b"], []],
        }
    )


@pytest.fixture
def t2(table_factory) -> ir.Table:
    return table_factory(
        {
            "record_id": [90, 91, 92, 93],
            "int": [2, 4, None, 6],
            "letter": ["b", "c", "d", None],
            "array": [["b"], ["c"], ["d"], None],
        }
    )


@dataclasses.dataclass
class ToShape:
    forward: Callable[[ir.Value], ir.Value]
    revert: Callable[[ir.Value], ir.Value]

    def call(self, f: Callable, *args, **kwargs) -> ir.Value:
        first, *rest = args
        args = (self.forward(first), *rest)
        result = f(*args, **kwargs)
        return self.revert(result)


@pytest.fixture(params=["scalar", "column"])
def to_shape(request) -> ToShape:
    """Fixture that allows you to test a function with both scalar and column inputs.

    Say you had some function ``add_one(x: ir.Value) -> ir.Value``.
    You already have a test like

    ```python
    inp = literal(1)
    result = add_one(inp)
    assert result.execute() == 2
    ```

    You can use this fixture to test add_one with a column input:

    ```
    inp = literal(1)
    result = to_shape.revert(add_one(to_shape.forward(inp)))
    assert result.execute() == 2
    ```

    Or, to do it in one step:

    ```
    inp = literal(1)
    result = to_shape.call(add_one, inp)
    assert result.execute() == 2
    ```
    """
    if request.param == "scalar":
        return ToShape(
            forward=lambda x: x,
            revert=lambda x: x,
        )
    elif request.param == "column":
        return ToShape(
            forward=lambda x: ibis.array([x]).unnest(),
            revert=lambda x: x.as_scalar(),
        )
    else:
        assert False


@pytest.fixture(autouse=True)
def set_env(monkeypatch):
    # disable color for doctests so we don't have to include escape codes in docstrings
    monkeypatch.setitem(os.environ, "NO_COLOR", "1")
    # Explicitly set the column width to be as large as needed
    monkeypatch.setitem(os.environ, "COLUMNS", "88")
    # reset interactive mode to False for doctests
    starting_opts = ibis.options
    ibis.options.repr.interactive.max_columns = 1000
    ibis.options.repr.interactive.max_depth = 3
    yield
    ibis.options = starting_opts


def download_test_data() -> ir.Table:
    # download test data from https://github.com/NickCrews/apoc-data/releases/tag/20240717-111158
    URL_TEMPLATE = "https://github.com/NickCrews/apoc-data/releases/download/20240717-111158/income_{year}.csv"
    conn = ibis.duckdb.connect()
    sub_tables = [
        conn.read_csv(
            (URL_TEMPLATE.format(year=year) for year in range(2011, 2024)),
            all_varchar=True,
        )
    ]
    t = ibis.union(*sub_tables)
    t = t.select(
        full_address=_.Address
        + ", "
        + _.City
        + ", "
        + _.State
        + ", "
        + _.Zip
        + ", "
        + _.Country
    )
    return t


@pytest.fixture
def addresses_1M(backend: ibis.BaseBackend) -> ir.Table:
    pq = DATA_DIR / "apoc_addresses_1M.parquet"
    if not pq.exists():
        download_test_data().to_parquet(pq)
    t = backend.read_parquet(pq)
    t = t.cache()  # ensure in memory, we don't want to benchmark disk IO
    return t



from __future__ import annotations

import ibis


class TableWrapper(ibis.Table):
    """A wrapper around an ibis Table that allows you to access its attributes.

    This allows the user to pretend that they are working with an ibis Table:

    - `isinstance(obj, ibis.Table)` is True
    - users can call all normal methods `t.distinct()`
    - can access columns, eg `t.column_name` or `t["column_name"]`

    but it also allows use as developers to extend the functionality of the table,
    eg adding custom methods or attributes.
    """

    __wrapped__: ibis.Table
    """The underlying Table that this is a proxy for."""

    def __init__(self, wrapped: ibis.Table, /) -> None:
        object.__setattr__(self, "__wrapped__", wrapped)

    def __getattr__(self, key: str) -> ibis.Column:
        return getattr(self.__wrapped__, key)


class StructWrapper(ibis.ir.StructValue):
    """A wrapper around an ibis StructValue that allows you to access its attributes.

    This allows the user to pretend that they are working with an ibis StructValue:

    - `isinstance(obj, ibis.ir.StructValue)` is True
    - users can call all normal methods `v.isnull()`
    - can access fields, eg `v.my_field` or `v["my_field"]`

    but it also allows use as developers to extend the functionality of the struct,
    eg adding custom methods or attributes.
    """

    __slots__ = ("__wrapped__",)
    __wrapped__: ibis.ir.StructValue
    """The underlying StructValue that this is a proxy for."""

    def __init__(self, wrapped: ibis.ir.StructValue, /) -> None:
        object.__setattr__(self, "__wrapped__", wrapped)

    def __getattr__(self, name: str, /) -> ibis.Value:
        return getattr(self.__wrapped__, name)



from __future__ import annotations

import ibis
from ibis import _
import pytest

from mismo.types import Updates


@pytest.fixture
def updates(table_factory):
    before = table_factory(
        {
            "id": [1, 2, 3, 4],
            "name": ["Alice", None, "Charlie", "David"],
            "age": [10, 20, 30, 40],
        }
    )
    after = table_factory(
        {
            "id": [1, 2, 3, 4],
            "name": ["Alice", None, "Charlie", None],
            "age": [10, 20, 99, 99],
        }
    )
    return Updates.from_tables(before, after, join_on="id")


def test_different_schemas(table_factory):
    before = table_factory({"id": [1, 2, 3]}, schema={"id": "int64"})
    after = table_factory({"id": [1, 2, 3]}, schema={"id": "!int64"})
    with pytest.raises(ValueError):
        Updates.from_tables(before, after, join_on="id")
    u = Updates.from_tables(before, after, join_on="id", check_schemas="names")
    assert dict(u.before().schema()) == dict(ibis.schema({"id": "int64"}))
    assert dict(u.after().schema()) == dict(ibis.schema({"id": "!int64"}))


def test_from_tables_no_join(table_factory):
    before = table_factory({"id": [1, 2]}, schema={"id": "int64"})
    after = table_factory({"id": [3, 4]}, schema={"id": "!int64"})
    u = Updates.from_tables(before, after, join_on=False, check_schemas="names")
    assert u.count().execute() == 0
    assert u.schema() == ibis.schema({"id": "struct<before: int64, after: !int64>"})


def test_any_different(updates: Updates):
    # actual = updates.filter(updates.filters.any_changed())
    any_changed = ibis.or_(*[updates[col].is_changed() for col in updates.columns])
    actual = updates.filter(any_changed)
    actual_ids = set(actual.after().id.execute())
    expected_ids = {3, 4}
    assert actual_ids == expected_ids


def test_all_different(updates: Updates):
    all_different = (updates[col].is_changed() for col in updates.columns)
    actual = updates.filter(*all_different)
    actual_ids = set(actual.after().id.execute())
    # the ids are the same lol
    expected_ids = set()
    assert actual_ids == expected_ids


def test_filter(updates: Updates):
    f = updates.filter(_.id.before == 3)
    assert isinstance(f, Updates)
    assert f.count().execute() == 1


def test_cache(updates: Updates):
    f = updates.cache()
    assert isinstance(f, Updates)
    assert (f.execute() == updates.execute()).all().all()



from __future__ import annotations

import numpy as np
import pytest

from mismo import text


@pytest.mark.parametrize(
    "input, expected",
    [
        pytest.param("hello", ["HL", "HL"], id="hello"),
        pytest.param("world", ["ARLT", "FRLT"], id="world"),
        pytest.param("hello world", ["HLRLT", "HLRLT"], id="hello world"),
        pytest.param("catherine", ["K0RN", "KTRN"], id="catherine"),
        pytest.param("", ["", ""], id="empty"),
        pytest.param(None, None, id="empty"),
    ],
)
def test_double_metaphone(input, expected):
    result = text.double_metaphone(input).execute()
    assert expected == result


@pytest.mark.parametrize(
    "string1,string2,expected",
    [
        ("foo", "foo", 1),
        ("bare", "bone", 0.5),
        ("baz", "def", 0),
        ("", "", np.nan),
        (None, None, np.nan),
    ],
)
def test_levenshtein_ratio(string1, string2, expected):
    result = text.levenshtein_ratio(string1, string2).execute()
    if np.isnan(expected):
        assert np.isnan(result)
    else:
        assert expected == result


@pytest.mark.parametrize(
    "string1,string2,expected",
    [
        ("foo", "foo", 1),
        ("foo", "food", 0.942),
        ("bar", "bim", 0.5555),
        ("a", "", 0),
        ("", "", 0),
        ("foo", None, np.nan),
        (None, None, np.nan),
    ],
)
def test_jaro_winkler_similarity(string1, string2, expected):
    result = text.jaro_winkler_similarity(string1, string2).execute()
    if np.isnan(expected):
        assert np.isnan(result)
    else:
        assert result == pytest.approx(expected, 0.001)


@pytest.mark.parametrize(
    "string1,string2,expected",
    [
        ("foo", "foo", 1),
        ("foo", "food", 0.916),
        ("bar", "bim", 0.5555),
        ("a", "", 0),
        ("", "", 0),
        ("foo", None, np.nan),
        (None, None, np.nan),
    ],
)
def test_jaro_similarity(string1, string2, expected):
    result = text.jaro_similarity(string1, string2).execute()
    if np.isnan(expected):
        assert np.isnan(result)
    else:
        assert result == pytest.approx(expected, 0.001)



from __future__ import annotations

from functools import cached_property

import ibis
from ibis import _
from ibis.expr import types as ir

from mismo import _util


class Factorizer:
    """Encode a column as integers.

    Use this class if you have some non-integer column, but you want to use
    integer codes to do some operation, and then restore the original column.
    """

    # We use the default of int64 because spark doesn't support unsigned ints
    def __init__(self, t: ir.Table, column: str, *, dtype: str = "int64") -> None:
        """Create a Factorizer from column of values.

        If the input column is already an integer, the Factorizer is a no-op.

        Parameters
        ----------
        t :
            The table containing the column to encode.
        column :
            The name of the column to encode.
        dtype:
            The datatype to encode to.
        """
        self.t = t
        self.column = column
        self._int_column = _util.unique_name()
        self.dtype = ibis.dtype(dtype)
        # The input column is already the right dtype, so the mapping
        # is simply the identity function. This is an optimization.
        self._noop = t[column].type() == self.dtype

    def encode(
        self,
        t: ir.Table | None = None,
        src: str | None = None,
        dst: str | None = None,
        verify: bool = True,
    ) -> ir.Table:
        """Encode a column to integer codes.

        Parameters
        ----------
        t :
            The table to encode. If None, use the table passed to the constructor.
        src :a
            The name of the  column to encode.
            If None, use the column passed to the constructor.
        dst :
            The name of the column to create. If None, overwrite the `src` column.
        verify :
            If True, raise an error if the column contains values that are not
            in the original column.
        """
        if t is None:
            t = self.t
        if src is None:
            src = self.column
        if dst is None:
            dst = src
        if self._noop:
            return t.mutate(**{dst: _[src]})

        if t[src] is self.t[self.column]:
            return self._augmented.mutate(**{dst: _[self._int_column]}).drop(
                self._int_column
            )

        if verify:
            if (
                self._mapping_count == 0
                and t.count().execute() > 0
                and t[src].notin(self._mapping.original).any().execute()
            ):
                raise ValueError(
                    f"Column {src} contains values that are not in the original column"
                )

        assert self._int_column not in t.columns
        joined = ibis.join(
            t, self._mapping, t[src] == self._mapping.original, how="left"
        )
        if src == "original":
            joined = joined.drop("original_right")
        joined = joined.drop("original")
        result = joined.mutate(**{dst: _[self._int_column]}).drop(self._int_column)
        return result

    def decode(
        self, t: ir.Table, src: str, dst: str | None = None, verify: bool = True
    ) -> ir.Table:
        """Decode a column of integer codes back to the original values.

        Parameters
        ----------
        t :
            The table to decode.
        src :
            The name of the column to decode.
        dst :
            The name of the column to create. If None, overwrite the `src` column.
        verify :
            If True, raise an error if the column contains codes that are not
            froms the original column.
        """
        if dst is None:
            dst = src
        if self._noop:
            return t.mutate(**{dst: _[src]})

        if verify:
            if (
                self._mapping_count == 0
                and t.count().execute() > 0
                and t[src].notin(self._mapping[self._int_column]).any().execute()
            ):
                raise ValueError(
                    f"Column {src} contains codes that are not from the original column"
                )

        assert self._int_column not in t.columns
        joined = ibis.join(
            t, self._mapping, t[src] == self._mapping[self._int_column], how="left"
        )
        orig = "original" if src != "original" else "original_right"
        result = joined.mutate(**{dst: _[orig]}).drop(orig, self._int_column)
        return result

    @cached_property
    def _augmented(self) -> ir.Table:
        return self.t.mutate(
            _util.group_id(self.column, dtype=self.dtype).name(self._int_column)
        )

    @cached_property
    def _mapping(self) -> ir.Table:
        return self._augmented.select(
            self._int_column, original=_[self.column]
        ).distinct()

    @cached_property
    def _mapping_count(self) -> int:
        return self._mapping.count().execute()



from __future__ import annotations

import ibis
import numpy as np
import pytest

from mismo.fs import ComparerWeights, LevelWeights, Weights


def test_level_weights_basic():
    lw = LevelWeights(name="close", m=0.5, u=0.05)
    assert lw.name == "close"
    assert lw.m == 0.5
    assert lw.u == 0.05
    assert lw.odds == 10
    assert lw.log_odds == 1
    assert lw == LevelWeights(name="close", m=0.5, u=0.05)
    assert lw != LevelWeights(name="oops", m=0.5, u=0.05)
    assert lw != 999
    assert repr(lw) == "LevelWeights(name=close, m=0.5, u=0.05)"

    # Can't set attributes
    with pytest.raises(AttributeError):
        lw.name = "a"
    with pytest.raises(AttributeError):
        lw.m = 0
    with pytest.raises(AttributeError):
        lw.u = 0
    with pytest.raises(AttributeError):
        lw.odds = 0


def test_level_weights_zero():
    lw = LevelWeights(name="close", m=0, u=0.6)
    assert lw.odds == 0
    assert lw.log_odds == float("-inf")


def test_level_weights_inf():
    lw = LevelWeights(name="close", m=0.5, u=0)
    assert lw.odds == float("inf")
    assert lw.log_odds == float("inf")


def test_level_weights_eq():
    lw1 = LevelWeights(name="close", m=0.5, u=0.05)
    lw2 = LevelWeights(name="close", m=0.5, u=0.05)
    lw3 = LevelWeights(name="close", m=0.5, u=0.05001)
    lw4 = LevelWeights(name="close", m=0.5, u=0.05000000001)
    assert lw1 == lw2
    assert lw1 != 999
    assert lw1 != lw3
    assert lw1 == lw4


def test_comparer_weights_basic():
    close_lw = LevelWeights(name="close", m=0.3, u=0.1)
    exact_lw = LevelWeights(name="exact", m=0.6, u=0.2)
    else_lw = LevelWeights(name="else", m=0.1, u=0.7)
    cw = ComparerWeights(
        name="address",
        level_weights=[close_lw, exact_lw, else_lw],
    )
    assert cw.name == "address"
    assert "close" in cw
    assert "exact" in cw
    assert "else" in cw
    assert "oops" not in cw
    assert 0 in cw
    assert 1 in cw
    assert 2 in cw
    assert 3 not in cw
    assert -1 in cw
    assert -2 in cw
    assert -3 in cw
    assert -4 not in cw

    assert len(cw) == 3
    assert cw["close"] == close_lw
    assert cw["exact"] == exact_lw
    assert cw["else"] == else_lw
    assert cw[0] == close_lw
    assert cw[1] == exact_lw
    assert cw[2] == else_lw
    assert cw[-1] == else_lw
    assert cw[-2] == exact_lw
    assert cw[-3] == close_lw
    with pytest.raises(KeyError):
        cw["oops"]
    with pytest.raises(KeyError):
        cw[3]
    with pytest.raises(KeyError):
        cw[-4]
    with pytest.raises(KeyError):
        cw[None]
    assert cw[:2] == (close_lw, exact_lw)
    assert cw[1:] == (exact_lw, else_lw)

    assert tuple(cw) == (close_lw, exact_lw, else_lw)
    assert list(cw) == [close_lw, exact_lw, else_lw]

    assert "address" in repr(cw)
    assert "close" in repr(cw)
    assert "exact" in repr(cw)
    assert "else" in repr(cw)


def test_comparer_weights_eq():
    close_lw = LevelWeights(name="close", m=0.3, u=0.1)
    exact_lw = LevelWeights(name="exact", m=0.6, u=0.2)
    cw1 = ComparerWeights(name="address", level_weights=[close_lw, exact_lw])
    cw2 = ComparerWeights(name="address", level_weights=[close_lw, exact_lw])
    cw3 = ComparerWeights(name="other", level_weights=[close_lw, exact_lw])
    cw4 = ComparerWeights(name="address", level_weights=[exact_lw, close_lw])
    assert cw1 == cw2
    assert cw1 != cw3
    assert cw1 != cw4
    assert cw1 != 999


def test_comparer_weights_odds():
    close_lw = LevelWeights(name="close", m=0.1, u=0.01)
    exact_lw = LevelWeights(name="exact", m=0.6, u=0.3)
    cw = ComparerWeights(name="address", level_weights=[close_lw, exact_lw])
    assert cw.odds("close") == 10
    assert cw.odds(0) == 10
    assert cw.odds("exact") == 2
    assert cw.odds(1) == 2

    with pytest.raises(KeyError):
        cw.odds("oops")
    with pytest.raises(KeyError):
        cw.odds(3)

    # When given an ibis expression, an ibis expression is returned
    assert cw.odds(ibis.literal("close")).execute() == 10
    assert cw.odds(ibis.literal(0)).execute() == 10
    # we can't have guardrails for giving bad levels :(
    assert np.isnan(cw.odds(ibis.literal("oops")).execute())
    assert np.isnan(cw.odds(ibis.literal(5)).execute())


def test_weights_serde(tmp_path):
    weights = Weights(
        [
            ComparerWeights(
                name="name",
                level_weights=[
                    LevelWeights(name="close", m=0.5, u=0.1),
                    LevelWeights(name="exact", m=0.6, u=0.9),
                ],
            ),
            ComparerWeights(
                name="address",
                level_weights=[
                    LevelWeights(name="within 10 km", m=0.5, u=0.1),
                    LevelWeights(name="street match", m=0.6, u=0.9),
                ],
            ),
        ]
    )
    d = weights.to_json(tmp_path / "weights.json")
    assert isinstance(d, dict)
    weights2 = Weights.from_json(tmp_path / "weights.json")
    weights3 = Weights.from_json(d)
    assert weights == weights2
    assert weights == weights3



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

import ibis
from ibis import _
from ibis.expr import types as ir

from mismo._util import bind_one
from mismo.arrays import array_choice
from mismo.linker import _common

if TYPE_CHECKING:
    import altair as alt


def minhash_lsh_keys(
    terms: ir.ArrayValue, *, band_size: int, n_bands: int
) -> ir.ArrayValue:
    """Create LSH keys from sets of terms."""
    # Many different flavors of how to implement minhash LSH,
    # I chose one based on
    # https://dl.acm.org/doi/10.1145/3511808.3557631
    bands = [array_choice(terms, band_size).sort() for _ in range(n_bands)]
    result = ibis.array(bands).filter(lambda x: x.length() > 0)

    def _hash(x):
        # NOTE: returns int64 in ibis 9.1.0 and later but uint64 in earlier versions
        if ibis.__version__ >= "9.1.0":
            # in later ibis versions, there is some post-processing
            # so the uint64s that duckdb returns are properly cast to int64s
            return x.hash()
        else:
            # in earlier versions, ibis *says* it returns int64s,
            # but duckdb actually returns uint64s, which
            # crashes things down the line without this cast.
            # See https://github.com/ibis-project/ibis/issues/9135
            # and https://github.com/NickCrews/mismo/issues/45
            return x.hash().cast("uint64")

    result = result.map(_hash)
    result = result.unique()
    return result


class MinhashLshLinker(_common.Linker):
    """A [Linker][mismo.Linker] that uses Minhash LSH to block record pairs that have high Jaccard similarity.

    See [the how-to guide](../howto/lsh.ipynb) for more information.
    """  # noqa: E501

    def __init__(
        self,
        *,
        terms_column: str,
        band_size: int,
        n_bands: int,
        keys_column: str = "{terms_column}_lsh_keys",
    ) -> None:
        """Make a Minhash LSH blocker.

        Parameters
        ----------
        terms_column :
            The column that holds the terms to compare.
        band_size :
            The number of terms in each band.
            See [plot_lsh_curves][mismo.block.plot_lsh_curves] for guidance.
        n_bands :
            The number of bands.
            See [plot_lsh_curves][mismo.block.plot_lsh_curves] for guidance.
        keys_column :
            The name of the column that will hold the LSH keys.
        """
        self.terms_column = terms_column
        self.band_size = band_size
        self.n_bands = n_bands
        self.keys_column = keys_column

    def __call__(self, left: ir.Table, right: ir.Table) -> ir.Table:
        """Block two tables using Minhash LSH."""
        left_terms = bind_one(left, self.terms_column)
        right_terms = bind_one(right, self.terms_column)
        keys_name = self.keys_column.format(terms_column=left_terms.get_name())
        left = left.mutate(
            minhash_lsh_keys(
                left_terms, band_size=self.band_size, n_bands=self.n_bands
            ).name(keys_name)
        )
        right = right.mutate(
            minhash_lsh_keys(
                right_terms, band_size=self.band_size, n_bands=self.n_bands
            ).name(keys_name)
        )
        left = left.cache()
        right = right.cache()
        # kb = KeyBlocker(_[keys_name].unnest())
        # return kb(left, right, task=task)


def p_blocked(jaccard: float, band_size: int, n_bands: int) -> float:
    return 1 - (1 - jaccard**band_size) ** n_bands


def plot_lsh_curves(
    band_params: Iterable[tuple[int, int]] | None = None,
) -> alt.Chart:
    """
    Plot performance of Minhash LSH for different band sizes and numbers of bands.

    MinhashLSH is a probabilistic method: The probability `P` that a pair of records
    will be blocked is a function of their jaccard similarity `J`.
    It is in a shape like a logistic curve, a rounded step function:

    ```
        ---------------------------------
    1.0 |                         x x x |
        |                       x       |
        |                      x        |
    P   |                      x        |
        |                      x        |
        |                      x        |
        |                     x         |
    0.0 | x x x x x x x x x x           |
        ---------------------------------
         0.0         Jaccard         1.0
    ```

    ![](../assets/lsh_curves.png)

    The shape of this curve is determined by the band size and number of bands.
    Use this function to choose the best band size and number of bands for your
    use case. Note that the runtime of the Minhash LSH blocker runs in
    `O(band_size * n_bands * n_records)` time, so you want to choose as small
    of band size and number of bands as possible.

    Based on [this PyData talk](https://youtu.be/n3dCcwWV4_k?si=Q9f4EuGtRE0xSyEg&t=1582)
    and the [accompanying code](https://github.com/mattilyra/LSH/blob/a57069bfb70f4b620d47931f81966b5a73c1b480/examples/Introduction.ipynb)

    Parameters
    ----------
    band_params :
        List of (band size, number of bands) to plot. If not provided, defaults to
        a reasonable selection to show the various curves.

    Returns
    -------
    The plot of the LSH curve.
    """
    import altair as alt

    if band_params is None:
        band_params = [
            (2, 10),
            (2, 25),
            (2, 50),
            (2, 100),
            (5, 20),
            (5, 40),
            (10, 10),
            (10, 20),
            (10, 50),
            (20, 5),
            (20, 10),
            (50, 2),
            (50, 4),
        ]
    t = ibis.memtable(data=band_params, columns=["band_size", "n_bands"])
    t = t.mutate(jaccard=ibis.range(51).unnest().cast(float) / 50)
    t = t.mutate(
        label="(" + t.band_size.cast(str) + ", " + t.n_bands.cast(str) + ")",
        pr=p_blocked(_.jaccard, _.band_size, _.n_bands),
    )
    chart = (
        alt.Chart(
            t.to_pandas(),
            title="Probability of LSH blocking a pair given a Jaccard similarity",
            width=400,
            height=400,
        )
        .mark_line(strokeWidth=2, point=True)
        .encode(
            x="jaccard:Q",
            y="pr:Q",
            color=alt.Color(
                "label:N",
                title="Band size, Number of bands",
                sort=alt.EncodingSortField("band_size"),
            ),
            tooltip=["band_size:Q", "n_bands:Q", "jaccard:Q", "pr:Q"],
        )
    )
    return chart



from __future__ import annotations

import mismo
from mismo.lib.name import NameDimension, NameMatchLevel
from mismo.tests.util import assert_tables_equal


def test_name_dimension(name_table, table_factory):
    dim = NameDimension("name")
    name_table = dim.prepare_for_fast_linking(name_table)
    name_table = dim.prepare_for_blocking(name_table)
    expected = table_factory(
        [
            ("a_anderson_phd", "a_anderson_phd", NameMatchLevel.EXACT.as_integer()),
            ("a_anderson_phd", "alice_anderson", NameMatchLevel.INITIALS.as_integer()),
            ("bob_baker", "robert_b_baker_jr", NameMatchLevel.NICKNAMES.as_integer()),
            (
                "charles_carter",
                "mr_charles_carter",
                NameMatchLevel.GIVEN_SURNAME.as_integer(),
            ),
            ("a_anderson_phd", "<null>", NameMatchLevel.NULL.as_integer()),
            ("alice_anderson", "charles_carter", NameMatchLevel.ELSE.as_integer()),
        ],
        schema={
            "record_id_l": "string",
            "record_id_r": "string",
            "name_compared": "int8",
        },
    )
    links = mismo.FullLinker(task="link")(name_table, name_table).links
    compared = dim.compare(links)
    compared = compared.semi_join(expected, ["record_id_l", "record_id_r"])
    compared = compared.cache()
    compared = compared.select(expected.columns)
    assert_tables_equal(expected, compared)



from __future__ import annotations

from ._address import AddressesDimension as AddressesDimension
from ._address import AddressesMatchLevel as AddressesMatchLevel
from ._address import match_level as match_level
from ._census import us_census_geocode as us_census_geocode
from ._latlon import CoordinateLinker as CoordinateLinker
from ._latlon import distance_km as distance_km
from ._postal import postal_fingerprint_address as postal_fingerprint_address
from ._postal import postal_parse_address as postal_parse_address
from ._regex_parse import parse_street1_re as parse_street1_re
from ._spacy import TaggedAddress as TaggedAddress
from ._spacy import spacy_tag_address as spacy_tag_address



from __future__ import annotations

from ibis.expr import types as ir
import pytest

import mismo


@pytest.mark.parametrize(
    "condition,is_slow",
    [
        pytest.param("letter", False, id="simple equijoin"),
        pytest.param(True, True, id="cross join"),
        pytest.param(
            lambda left, right, **_: left.letter.levenshtein(right.letter) < 2,
            True,
            id="levenshtein",
        ),
        pytest.param(
            lambda left, right, **_: (left.letter == right.letter)
            | (left.record_id == right.record_id),
            True,
            id="OR",
        ),
        pytest.param(
            lambda left, right, **_: (left.letter == right.letter)
            & (left.record_id == right.record_id),
            False,
            id="AND",
        ),
    ],
)
@pytest.mark.parametrize(
    "on_slow,result",
    [
        ("ignore", None),
        ("warn", mismo.exceptions.SlowJoinWarning),
        ("error", mismo.exceptions.SlowJoinError),
    ],
)
def test_check_join_algorithm(
    t1: ir.Table, t2: ir.Table, condition, is_slow, on_slow, result
):
    def f():
        mismo.joins.check_join_algorithm(t1, t2, condition, on_slow=on_slow)

    if result is None:
        f()
    elif is_slow and result is mismo.exceptions.SlowJoinWarning:
        with pytest.warns(mismo.exceptions.SlowJoinWarning):
            f()
    elif is_slow and result is mismo.exceptions.SlowJoinError:
        with pytest.raises(mismo.exceptions.SlowJoinError):
            f()



from __future__ import annotations

from collections.abc import Iterable
import operator
from typing import (
    Any,
    Callable,
    Generator,
    Protocol,
    TypeAlias,
    Union,
    runtime_checkable,
)

import ibis
from ibis import Deferred
from ibis.common.deferred import BinaryOperator, Resolver, Variable

from mismo import _funcs, _util


@runtime_checkable
class ValueResolver(Protocol):
    """A callable, that given a Table, resolves to a single Value."""

    def __call__(self, t: ibis.Table, /) -> ibis.Value:
        """Given a Table, resolve to a single Value."""


class DeferredResolver(ValueResolver):
    def __init__(
        self,
        deferred: ibis.Deferred,
        name: str | None = None,
    ) -> None:
        if name is None:
            name = "_"
        self.deferred = deferred
        self.name = name

    def __call__(self, t: ibis.Table, /) -> ibis.Value:
        raw = self.deferred.resolve(**{self.name: t})
        return _util.bind_one(t, raw)

    def __repr__(self) -> str:
        if self.name == "_":
            return f"DeferredResolver({self.deferred!r})"
        return f"DeferredResolver({self.deferred!r}, name={self.name!r})"

    def __str__(self) -> str:
        return f"{self.deferred!r}"


class LiteralResolver(ValueResolver):
    """A resolver that always returns the literal value given."""

    def __init__(self, value: ibis.Value, /) -> None:
        self.value = value

    def __call__(self, t: ibis.Table, /) -> ibis.Value:
        return _util.bind_one(t, self.value)

    def __repr__(self) -> str:
        return f"LiteralResolver({self.value!r})"

    def __str__(self) -> str:
        return f"`{self.value!r}`"


class StrResolver(ValueResolver):
    def __init__(self, s: str, /) -> None:
        self.s = s

    def __call__(self, t: ibis.Table, /) -> ibis.Column:
        """Resolve a string to a column."""
        return t[self.s]

    def __repr__(self):
        return f"StrResolver({self.s!r})"

    def __str__(self):
        return f"`{self.s!r}`"


class FuncResolver(ValueResolver):
    def __init__(self, func: Callable[[ibis.Table], ibis.Value], /) -> None:
        self.func = func

    def __call__(self, t: ibis.Table, /) -> ibis.Value:
        return self.func(t)

    def __repr__(self):
        return f"FuncResolver({self.func!r})"


IntoValueResolver: TypeAlias = Union[
    ibis.Value,
    ibis.Deferred,
    str,
    Callable[[ibis.Table], "IntoValueResolver"],
]


def value_resolver(spec: IntoValueResolver) -> ValueResolver:
    """
    Given a spec, return a ValueResolver that resolves to a single column.
    """
    if isinstance(spec, ibis.Value):
        return LiteralResolver(spec)
    if isinstance(spec, ibis.Deferred):
        if variables_names(spec) == {"_"}:
            return DeferredResolver(spec, "_")
        raise ValueError(f"Deferred objects must represent a single column, got {spec}")
    if isinstance(spec, str):
        return StrResolver(spec)
    if _funcs.is_unary(spec):
        return FuncResolver(spec)
    raise ValueError(f"Cannot resolve {spec} to a single column.")


IntoKeyPairResolver: TypeAlias = Union[
    ibis.Value,
    ibis.Deferred,
    str,
    Callable[[ibis.Table], "IntoValueResolver"],
    Iterable[
        Union[
            ibis.Value,
            ibis.Deferred,
            str,
            Callable[[ibis.Table], "IntoValueResolver"],
        ]
    ],
]


def key_pair_resolver(spec: IntoKeyPairResolver) -> tuple[ValueResolver, ValueResolver]:
    if isinstance(spec, ibis.Value):
        return (value_resolver(spec), value_resolver(spec))
    if isinstance(spec, ibis.Deferred):
        if variables_names(spec) == {"_"}:
            return (value_resolver(spec), value_resolver(spec))
        raise ValueError(f"Deferred objects must represent a single column, got {spec}")
    if isinstance(spec, str):
        return (value_resolver(spec), value_resolver(spec))
    if _funcs.is_unary(spec):
        return (value_resolver(spec), value_resolver(spec))
    parts = _util.promote_list(spec)
    if len(parts) != 2:
        raise ValueError(
            (
                "Column spec, when an iterable, must be length 2.",
                f" got {len(parts)} from {spec}",
            )
        )
    keyl, keyr = parts
    if isinstance(keyl, Deferred):
        left_resolver = DeferredResolver(keyl, _get_name(keyl, {"left", "_"}))
    else:
        left_resolver = value_resolver(keyl)
    if isinstance(keyr, Deferred):
        right_resolver = DeferredResolver(keyr, _get_name(keyr, {"right", "_"}))
    else:
        right_resolver = value_resolver(keyr)

    return (left_resolver, right_resolver)


def _get_name(d: Deferred, allowed_names: set[str]) -> str:
    """Get the name of a Deferred, or raise an error if it is not in allowed_names."""
    actual_names = variables_names(d)
    if len(actual_names) != 1:
        raise ValueError(
            f"Expected a Deferred with name in {allowed_names}, got {actual_names} from {d}"  # noqa: E501
        )
    actual_name = actual_names.pop()
    if actual_name not in allowed_names:
        raise ValueError(
            f"Expected a Deferred with name in {allowed_names}, got {actual_name} from {d}"  # noqa: E501
        )
    return actual_name


def key_pair_resolvers(x) -> list[tuple[ValueResolver, ValueResolver]]:
    """
    Given a spec or iterable of specs, return a list of KeyPairResolvers.
    """
    if (deferred_resolvers := _parse_and_of_equals(x)) is not None:
        return deferred_resolvers  # ty:ignore[invalid-return-type]
    return [key_pair_resolver(spec) for spec in _util.promote_list(x)]


def _parse_and_of_equals(x) -> list[tuple[DeferredResolver, DeferredResolver]] | None:
    """Look for the special case of an `and` of equalities, like `left.col1 == right.col2 & left.col3 == right.col4 & ...`"""  # noqa: E501
    if not isinstance(x, ibis.Deferred):
        return None
    if variables_names(x) != {"left", "right"}:
        return None
    resolver = _to_resolver(x)
    return list(_traverse_deferred_resolvers(resolver))


class BadDeferredError(ValueError):
    pass


def _traverse_deferred_resolvers(
    resolver: Resolver,
) -> Generator[tuple[DeferredResolver, DeferredResolver], None, None]:
    if not isinstance(resolver, BinaryOperator):
        return
    if resolver.func == operator.and_:
        # Recurse on the left and right operands
        yield from _traverse_deferred_resolvers(resolver.left)
        yield from _traverse_deferred_resolvers(resolver.right)
    elif resolver.func == operator.eq:
        if variables_names(resolver.left) == {"left"} and variables_names(
            resolver.right
        ) == {"right"}:
            left_resolver = resolver.left
            right_resolver = resolver.right
        elif variables_names(resolver.left) == {"right"} and variables_names(
            resolver.right
        ) == {"left"}:
            left_resolver = resolver.right
            right_resolver = resolver.left
        else:
            raise BadDeferredError(
                f"Expected an equality between left and right, got {resolver}"
            )
        yield (
            DeferredResolver(Deferred(left_resolver), name="left"),
            DeferredResolver(Deferred(right_resolver), name="right"),
        )
    else:
        raise BadDeferredError(
            "Expected 1-N equality conditions combined with `and`"
            "(eg `left.col1 == right.col2 & left.col3 == right.col4 & ...`), "
            f"got {resolver.func} instead."
        )


@runtime_checkable
class HasResolver(Protocol):
    _resolver: Resolver


def _to_resolver(x: Resolver | HasResolver | Deferred) -> Resolver:
    if isinstance(x, Resolver):
        return x
    if isinstance(x, HasResolver):
        return x._resolver
    raise TypeError(
        "Expected a Resolver or something with `_resolver`,"
        f" got {type(x).__name__} instead."
    )


def variables_names(deferred: Deferred | Resolver) -> set[str]:
    """Get the names of all Variables in a Deferred.

    Examples
    --------
    >>> import ibis
    >>> import mismo

    The builtin deferred variable `ibis._` has the name "_".

    >>> sorted(variables_names(ibis._.foo + ibis._.fill_null(5)))
    ['_']

    The `mismo.left` and `mismo.right` variables have the names "left" and "right".

    >>> sorted(variables_names(mismo.left.foo + mismo.right.bar + ibis._.baz))
    ['_', 'left', 'right']

    We don't really use this anywhere in mismo, but just so you see how it works,
    you can create your own deferred variables with names.

    >>> from ibis.common.deferred import var
    >>> sorted(variables_names(ibis._.foo + var("bar").baz.fill_null(8)))
    ['_', 'bar']
    """
    resolver = _to_resolver(deferred)
    if not isinstance(resolver, Resolver):
        raise TypeError(
            f"Expected a Deferred or Resolver, got {type(deferred).__name__} instead."
        )
    names = set()
    for resolver in _traverse_resolver(resolver):
        if isinstance(resolver, Variable):
            names.add(resolver.name)
    return names


def _traverse_resolver(x: Any) -> Generator[Resolver, None, None]:
    if isinstance(x, Resolver):
        yield x
    if hasattr(x, "__slots__"):
        for slot in x.__slots__:
            child = getattr(x, slot)
            yield from _traverse_resolver(child)



from __future__ import annotations

import ibis

from mismo import text


def test_re_extract_struct():
    s = ibis.literal("2024-03-01")
    pattern = r"(?P<year>\d+)-(?P<month>\d+)-(?P<day>\d+)"
    result = text.re_extract_struct(s, pattern).execute()
    assert result == {"year": "2024", "month": "03", "day": "01"}


def test_re_extract_struct_case_insensitive():
    s = ibis.literal("Hello World")
    pattern = r"HELLO (?P<word>\w+)"

    result = text.re_extract_struct(s, pattern, case_insensitive=True).execute()
    assert result == {"word": "World"}

    result = text.re_extract_struct(s, pattern, case_insensitive=False).execute()
    assert result == {"word": ""}

    result = text.re_extract_struct(s, pattern).execute()
    assert result == {"word": ""}



from __future__ import annotations

import ibis
import pytest

import mismo
from mismo import _resolve


@pytest.fixture
def left():
    return ibis.table(
        {
            "username": "string",
            "address": "string",
            "age": "int64",
            "is_true": "boolean",
        },
        name="left",
    )


@pytest.fixture
def right():
    return ibis.table(
        {
            "USERNAME": "string",
            "address": "string",
            "age": "int64",
        },
        name="right",
    )


@pytest.mark.parametrize(
    "condition",
    [
        pytest.param(
            ibis.and_(
                mismo.left.username == mismo.right.USERNAME,
                mismo.left.address == mismo.right.address.upper(),
                mismo.left.age == mismo.right.age + 1,
            ),
            id="flat",
        ),
        pytest.param(
            ibis.and_(
                mismo.left.username == mismo.right.USERNAME,
                ibis.and_(
                    mismo.left.address == mismo.right.address.upper(),
                    mismo.left.age == mismo.right.age + 1,
                ),
            ),
            id="nested",
        ),
    ],
)
def test_key_pair_resolvers_happy(condition, left, right):
    resolvers = _resolve.key_pair_resolvers(condition)
    assert len(resolvers) == 3
    (r1l, r1r), (r2l, r2r), (r3l, r3r) = resolvers
    # assert isinstance(r1, _resolve.IndividualDeferredResolver)
    assert r1l(left).equals(left.username)
    assert r1r(right).equals(right.USERNAME)
    # assert isinstance(r2, _resolve.IndividualDeferredResolver)
    assert r2l(left).equals(left.address)
    assert r2r(right).equals(right.address.upper())
    # assert isinstance(r3, _resolve.IndividualDeferredResolver)
    assert r3l(left).equals(left.age)
    assert r3r(right).equals(right.age + 1)


def test_key_pair_resolvers_single(left, right):
    condition = mismo.left.username == mismo.right.USERNAME
    resolvers = _resolve.key_pair_resolvers(condition)
    assert len(resolvers) == 1
    ((rl, rr),) = resolvers
    # assert isinstance(r1, _resolve.IndividualDeferredResolver)
    assert rl(left).equals(left.username)
    assert rr(right).equals(right.USERNAME)


def test_key_pair_resolvers_multiple_equals(left, right):
    condition = mismo.left.is_true == (mismo.right.USERNAME == "foo")
    resolvers = _resolve.key_pair_resolvers(condition)
    assert len(resolvers) == 1
    ((r1l, r1r),) = resolvers
    # assert isinstance(r1, _resolve.IndividualDeferredResolver)
    assert r1l(left).equals(left.is_true)
    assert r1r(right).equals(right.USERNAME == "foo")


def test_key_pair_resolvers_ors():
    condition = ibis.or_(
        mismo.left.username == mismo.right.USERNAME,
        mismo.left.address == mismo.right.address.upper(),
    )
    with pytest.raises(ValueError):
        _resolve.key_pair_resolvers(condition)


def test_key_pair_resolvers_nonequality():
    condition = ibis.and_(
        mismo.left.username == mismo.right.USERNAME,
        mismo.left.age > mismo.right.age + 1,
    )
    with pytest.raises(ValueError):
        _resolve.key_pair_resolvers(condition)



from __future__ import annotations

from ibis.expr import types as ir

from mismo.lib.name import _clean, _compare


class NameDimension:
    """Prepares, blocks, and compares based on a human name.

    A name is a Struct of the type
    `struct<
        prefix: string,
        given: string,
        middle: string,
        surname: string,
        suffix: string,
        nickname: string,
    >`.
    """

    def __init__(
        self,
        column: str,
        *,
        column_normed: str = "{column}_normed",
        column_tokens: str = "{column}_tokens",
        column_compared: str = "{column}_compared",
    ) -> None:
        self.column = column
        self.column_normed = column_normed.format(column=column)
        self.column_tokens = column_tokens.format(column=column)
        self.column_compared = column_compared.format(column=column)
        self.comparer = _compare.NameComparer(
            self.column_normed + "_l",
            self.column_normed + "_r",
            result_column=self.column_compared,
        )

    def prepare_for_fast_linking(self, t: ir.Table) -> ir.Table:
        """Add columns with the normalized name.

        Parameters
        ----------
        t
            The table to prep.

        Returns
        -------
        t
            The prepped table.
        """
        t = t.mutate(_clean.normalize_name(t[self.column]).name(self.column_normed))
        return t

    def prepare_for_blocking(self, t: ir.Table) -> ir.Table:
        t = t.mutate(_clean.name_tokens(t[self.column_normed]).name(self.column_tokens))
        return t

    def compare(self, t: ir.Table) -> ir.Table:
        """Compare the left and right names.

        Parameters
        ----------
        t :
            The table to compare.

        Returns
        -------
        t :
            The compared table.
        """
        return t.mutate(self.comparer(t))



from __future__ import annotations

from ._blocker import NameBlocker as NameBlocker
from ._clean import NameStruct as NameStruct
from ._clean import normalize_name as normalize_name
from ._compare import NameComparer as NameComparer
from ._compare import NameMatchLevel as NameMatchLevel
from ._compare import equal_forgiving_typo as equal_forgiving_typo
from ._dimension import NameDimension as NameDimension
from ._nicknames import are_aliases as are_aliases
from ._nicknames import is_nickname_for as is_nickname_for
